# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
''' Time series diagnostics

These started life as ``tsdiffana.m`` - see
http://imaging.mrc-cbu.cam.ac.uk/imaging/DataDiagnostics

Oliver Josephs (FIL) gave me (MB) the idea of time-point to time-point
subtraction as a diagnostic for motion and other sudden image changes.
'''
from __future__ import absolute_import

import numpy as np

from ...io.api import as_image

from ...core.reference.coordinate_map import (io_axis_indices, drop_io_dim, AxisError)


def time_slice_diffs(arr, time_axis=-1, slice_axis=None):
    ''' Time-point to time-point differences over volumes and slices

    We think of the passed array as an image.  The image has a "time"
    dimension given by `time_axis` and a "slice" dimension, given by
    `slice_axis`, and one or more other dimensions.  In the case of imaging
    there will usually be two more dimensions (the dimensions defining the size
    of an image slice). A single slice in the time dimension we call a "volume".
    A single entry in `arr` is a "voxel".  For example, if `time_axis` == 0,
    then ``v = arr[0]`` would be the first volume in the series.  The volume
    ``v`` above has ``v.size`` voxels.  If, in addition, `slice_axis` == 1, then
    for the volume ``v`` (above) ``s = v[0]`` would be a "slice", with
    ``s.size`` voxels. These are obviously terms from neuroimaging.

    Parameters
    ----------
    arr : array_like
       Array over which to calculate time and slice differences.  We'll
       call this array an 'image' in this doc.
    time_axis : int, optional
       axis of `arr` that varies over time. Default is last
    slice_axis : None or int, optional
       axis of `arr` that varies over image slice.  None gives last non-time
       axis.

    Returns
    -------
    results : dict

        ``T`` is the number of time points (``arr.shape[time_axis]``)

        ``S`` is the number of slices (``arr.shape[slice_axis]``)

        ``v`` is the shape of a volume (``rollimg(arr, time_axis)[0].shape``)

        ``d2[t]`` is the volume of squared differences between voxels at
        time point ``t`` and time point ``t+1``

        `results` has keys:

        * 'volume_mean_diff2' : (T-1,) array
           array containing the mean (over voxels in volume) of the
           squared difference from one time point to the next
        * 'slice_mean_diff2' : (T-1, S) array
           giving the mean (over voxels in slice) of the difference from
           one time point to the next, one value per slice, per
           timepoint
        * 'volume_means' : (T,) array
           mean over voxels for each volume ``vol[t] for t in 0:T``
        * 'slice_diff2_max_vol' : v[:] array
           volume, of same shape as input time point volumes, where each slice
           is is the slice from ``d2[t]`` for t in 0:T-1, that has the largest
           variance across ``t``.  Thus each slice in the volume may well result
           from a different difference time point.
        * 'diff2_mean_vol`` : v[:] array
           volume with the mean of ``d2[t]`` across t for t in 0:T-1.

    Raises
    ------
    ValueError : if `time_axis` refers to same axis as `slice_axis`
    '''
    arr = np.asarray(arr)
    ndim = arr.ndim
    # roll time axis to 0, slice axis to 1 for convenience
    if time_axis < 0:
        time_axis += ndim
    if slice_axis is None:
        slice_axis = ndim-2 if time_axis == ndim-1 else ndim-1
    elif slice_axis < 0:
        slice_axis += ndim
    if time_axis == slice_axis:
        raise ValueError('Time axis refers to same axis as slice axis')
    arr = np.rollaxis(arr, time_axis)
    # we may have changed the position of slice_axis
    if time_axis > slice_axis:
        slice_axis += 1
    arr = np.rollaxis(arr, slice_axis, 1)
    # shapes of things
    shape = arr.shape
    T = shape[0]
    S = shape[1]
    vol_shape = shape[1:]
    # loop over time points to save memory
    volds = np.empty((T-1,))
    sliceds = np.empty((T-1,S))
    means = np.empty((T,))
    diff_mean_vol = np.zeros(vol_shape)
    slice_diff_max_vol = np.zeros(vol_shape)
    slice_diff_maxes = np.zeros(S)
    last_tp = arr[0]
    means[0] = last_tp.mean()
    for dtpi in range(0,T-1):
        tp = arr[dtpi+1] # shape vol_shape
        means[dtpi+1] = tp.mean()
        dtp_diff2 = (tp - last_tp)**2
        diff_mean_vol += dtp_diff2
        sliceds[dtpi] = dtp_diff2.reshape(S, -1).mean(-1)
        # check whether we have found a highest-diff slice
        sdmx_higher = sliceds[dtpi] > slice_diff_maxes
        if any(sdmx_higher):
            slice_diff_maxes[sdmx_higher] = sliceds[dtpi][sdmx_higher]
            slice_diff_max_vol[sdmx_higher] = dtp_diff2[sdmx_higher]
        last_tp = tp
    volds = sliceds.mean(1)
    diff_mean_vol /= (T-1)
    # roll vol shapes back to match input
    diff_mean_vol = np.rollaxis(diff_mean_vol, 0, slice_axis)
    slice_diff_max_vol = np.rollaxis(slice_diff_max_vol, 0, slice_axis)
    return {'volume_mean_diff2': volds,
            'slice_mean_diff2': sliceds,
            'volume_means': means,
            'diff2_mean_vol': diff_mean_vol,
            'slice_diff2_max_vol': slice_diff_max_vol}


def time_slice_diffs_image(img, time_axis='t', slice_axis='slice'):
    """ Time-point to time-point differences over volumes and slices of image

    Parameters
    ----------
    img : Image
        The image on which to perform time-point differences
    time_axis : str or int, optional
        Axis indexing time-points. Default is 't'. If `time_axis` is an integer,
        gives the index of the input (domain) axis of `img`. If `time_axis` is a str,
        can be an input (domain) name, or an output (range) name, that maps to
        an input (domain) name.
    slice_axis : str or int, optional
        Axis indexing MRI slices. If `slice_axis` is an integer, gives the
        index of the input (domain) axis of `img`. If `slice_axis` is a str,
        can be an input (domain) name, or an output (range) name, that maps to
        an input (domain) name.

    Returns
    -------
    results : dict

        `arr` refers to the array as loaded from `img`

        ``T`` is the number of time points (``img.shape[time_axis]``)

        ``S`` is the number of slices (``img.shape[slice_axis]``)

        ``v`` is the shape of a volume (``rollimg(img, time_axis)[0].shape``)

        ``d2[t]`` is the volume of squared differences between voxels at
        time point ``t`` and time point ``t+1``

        `results` has keys:

        * 'volume_mean_diff2' : (T-1,) array
           array containing the mean (over voxels in volume) of the
           squared difference from one time point to the next
        * 'slice_mean_diff2' : (T-1, S) array
           giving the mean (over voxels in slice) of the difference from
           one time point to the next, one value per slice, per
           timepoint
        * 'volume_means' : (T,) array
           mean over voxels for each volume ``vol[t] for t in 0:T``
        * 'slice_diff2_max_vol' : v[:] image
           image volume, of same shape as input time point volumes, where each
           slice is is the slice from ``d2[t]`` for t in 0:T-1, that has the
           largest variance across ``t``.  Thus each slice in the volume may
           well result from a different difference time point.
        * 'diff2_mean_vol`` : v[:] image
           image volume with the mean of ``d2[t]`` across t for t in 0:T-1.
    """
    img = as_image(img)
    img_class = img.__class__
    time_in_ax, time_out_ax = io_axis_indices(img.coordmap, time_axis)
    if None in (time_in_ax, time_out_ax):
        raise AxisError('Cannot identify matching input output axes with "%s"'
                        % time_axis)
    slice_in_ax, slice_out_ax = io_axis_indices(img.coordmap, slice_axis)
    if None in (slice_in_ax, slice_out_ax):
        raise AxisError('Cannot identify matching input output axes with "%s"'
                        % slice_axis)
    vol_coordmap = drop_io_dim(img.coordmap, time_axis)
    results = time_slice_diffs(img.get_data(), time_in_ax, slice_in_ax)
    for key in ('slice_diff2_max_vol', 'diff2_mean_vol'):
        vol = img_class(results[key], vol_coordmap)
        results[key] = vol
    return results



from __future__ import absolute_import
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from numpy.testing import assert_equal, assert_almost_equal
import numpy as np

from .. import onesample


def test_onesample_stat():
    dx, dy, dz = 3, 4, 2
    nvox = dx*dy*dz
    nsub = 12
    # Make surrogate data 
    aux = np.arange(nvox)
    x = np.reshape(aux.repeat(nsub), [dx, dy, dz, nsub])
    # Gold standard 
    y_target = np.inf * np.ones(nvox)
    y_target[0] = 0.0
    # Test: input C-contiguous, data owner, axis=3
    y = onesample.stat(x, axis=3).reshape(nvox)
    assert_equal(y, y_target)
    # Test: input F-contiguous, not owner, axis=0 
    y = onesample.stat(x.T, axis=0).reshape(nvox)
    assert_equal(y, y_target)
    # Test: input C-contiguous, data owner, axis=0
    xT = x.T.copy()
    y = onesample.stat(xT, axis=0).reshape(nvox)
    assert_equal(y, y_target)
    

    
if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])



""" Testing timefuncs module
"""

from __future__ import division, print_function, absolute_import

import numpy as np

from numpy.testing import (assert_almost_equal,
                           assert_array_equal)

from nose.tools import (assert_true, assert_false, assert_raises,
                        assert_equal, assert_not_equal)


from .. import timefuncs as tf


def test_ascending():
    tr = 2.
    for func in (tf.st_01234, tf.ascending):
        for n_slices in (10, 11):
            assert_almost_equal(
                func(n_slices, tr),
                np.arange(n_slices) / n_slices * tr)
        assert_array_equal(
            np.argsort(func(5, 1)), [0, 1, 2, 3, 4])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_descending():
    tr = 2.
    for func in (tf.st_43210, tf.descending):
        for n_slices in (10, 11):
            assert_almost_equal(
                func(n_slices, tr),
                np.arange(n_slices-1, -1, -1) / n_slices * tr)
        assert_array_equal(
            np.argsort(func(5, 1)), [4, 3, 2, 1, 0])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_asc_alt_2():
    tr = 2.
    for func in (tf.st_02413, tf.asc_alt_2):
        assert_almost_equal(
            func(10, tr) / tr * 10,
            [0, 5, 1, 6, 2, 7, 3, 8, 4, 9])
        assert_almost_equal(
            func(11, tr) / tr * 11,
            [0, 6, 1, 7, 2, 8, 3, 9, 4, 10, 5])
        assert_array_equal(
            np.argsort(func(5, 1)), [0, 2, 4, 1, 3])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_desc_alt_2():
    tr = 2.
    for func in (tf.st_42031, tf.desc_alt_2):
        assert_almost_equal(
            func(10, tr) / tr * 10,
            [9, 4, 8, 3, 7, 2, 6, 1, 5, 0])
        assert_almost_equal(
            func(11, tr) / tr * 11,
            [5, 10, 4, 9, 3, 8, 2, 7, 1, 6, 0])
        assert_array_equal(
            np.argsort(func(5, 1)), [4, 2, 0, 3, 1])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_asc_alt_2_1():
    tr = 2.
    for func in (tf.st_13024, tf.asc_alt_2_1):
        assert_almost_equal(
            func(10, tr) / tr * 10,
            [5, 0, 6, 1, 7, 2, 8, 3, 9, 4])
        assert_almost_equal(
            func(11, tr) / tr * 11,
            [5, 0, 6, 1, 7, 2, 8, 3, 9, 4, 10])
        assert_array_equal(
            np.argsort(func(5, 1)), [1, 3, 0, 2, 4])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_asc_alt_siemens():
    tr = 2.
    for func in (tf.st_odd0_even1, tf.asc_alt_siemens):
        assert_almost_equal(
            func(10, tr) / tr * 10,
            [5, 0, 6, 1, 7, 2, 8, 3, 9, 4])
        assert_almost_equal(
            func(11, tr) / tr * 11,
            [0, 6, 1, 7, 2, 8, 3, 9, 4, 10, 5])
        assert_array_equal(
            np.argsort(func(5, 1)), [0, 2, 4, 1, 3])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_asc_alt_half():
    tr = 2.
    for func in (tf.st_03142, tf.asc_alt_half):
        assert_almost_equal(
            func(10, tr) / tr * 10,
            [0, 2, 4, 6, 8, 1, 3, 5, 7, 9])
        assert_almost_equal(
            func(11, tr) / tr * 11,
            [0, 2, 4, 6, 8, 10, 1, 3, 5, 7, 9])
        assert_array_equal(
            np.argsort(func(5, 1)), [0, 3, 1, 4, 2])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_desc_alt_half():
    tr = 2.
    for func in (tf.st_41302, tf.desc_alt_half):
        assert_almost_equal(
            func(10, tr) / tr * 10,
            [9, 7, 5, 3, 1, 8, 6, 4, 2, 0])
        assert_almost_equal(
            func(11, tr) / tr * 11,
            [9, 7, 5, 3, 1, 10, 8, 6, 4, 2, 0])
        assert_array_equal(
            np.argsort(func(5, 1)), [4, 1, 3, 0, 2])
        assert_equal(tf.SLICETIME_FUNCTIONS[func.__name__], func)


def test_number_names():
    for func in (
        tf.st_01234,
        tf.st_43210,
        tf.st_02413,
        tf.st_42031,
        tf.st_13024,
        tf.st_03142,
        tf.st_41302):
        name = func.__name__
        assert_equal(tf.SLICETIME_FUNCTIONS[name], func)
        assert_equal(tf.SLICETIME_FUNCTIONS[name[3:]], func)



from __future__ import absolute_import
from __future__ import print_function

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    # We need this because libcstat.a is linked to lapack, which can
    # be a fortran library, and the linker needs this information.
    config = Configuration('group', parent_package, top_path)
    config.add_subpackage('tests')
    config.add_extension(
        'onesample',
        sources=['onesample.pyx'],
        libraries=['cstat']
        )
    config.add_extension(
        'twosample',
        sources=['twosample.pyx'],
        libraries=['cstat']
        )
    config.add_extension(
        'glm_twolevel',
        sources=['glm_twolevel.pyx'],
        libraries=['cstat']
        )
    return config


if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



from __future__ import absolute_import
from __future__ import print_function
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('bindings', parent_package, top_path)
    config.add_subpackage('tests')
    config.add_subpackage('benchmarks')
    config.add_extension('linalg', sources=['linalg.pyx'],
                            libraries=['cstat'])
    config.add_extension('array', sources=['array.pyx'],
                            libraries=['cstat'])
    config.add_extension('wrapper', sources=['wrapper.pyx'],
                            libraries=['cstat'])
    return config


if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



""" Testing interpolation module
"""

import numpy as np

from nipy.core.api import Image, vox2mni

from ..interpolation import ImageInterpolator

from numpy.testing import (assert_almost_equal,
                           assert_array_equal)

from nose.tools import (assert_true, assert_false, assert_raises,
                        assert_equal, assert_not_equal)


def test_interpolator():
    arr = np.arange(24).reshape((2, 3, 4))
    coordmap =  vox2mni(np.eye(4))
    img = Image(arr, coordmap)
    # Interpolate off top right corner with different modes
    interp = ImageInterpolator(img, mode='nearest')
    assert_almost_equal(interp.evaluate([0, 0, 4]), arr[0, 0, -1])
    interp = ImageInterpolator(img, mode='constant', cval=0)
    assert_array_equal(interp.evaluate([0, 0, 4]), 0)
    interp = ImageInterpolator(img, mode='constant', cval=1)
    assert_array_equal(interp.evaluate([0, 0, 4]), 1)



# Init for slicetiming subpackage
""" Slicetiming subpackage

The standard nipy method of slice timing is implemented in
:mod:`nipy.algorithms.registration.groupwise_registration`.

This subpackage is a placeholder for other slice timing methods, and for utility
functions for slice timing
"""



from __future__ import absolute_import, print_function
#!/usr/bin/env python

def configuration(parent_package='',top_path=None):

    from numpy.distutils.misc_util import Configuration

    config = Configuration('clustering', parent_package, top_path)
    config.add_subpackage('tests')

    return config


if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Output a summary table for neuroimaging files (resolution, dimensionality, etc.)
"""

from nibabel.cmdline.ls import main


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

"""
A scripting wrapper around 4D registration (SpaceTimeRealign)
"""
from __future__ import absolute_import

import os
import os.path as op
import numpy as np
import numpy.linalg as npl

import nibabel as nib
from nibabel.filename_parser import splitext_addext
import nibabel.eulerangles as euler
from nibabel.optpkg import optional_package
matplotlib, HAVE_MPL, _ = optional_package('matplotlib')

from .groupwise_registration import SpaceTimeRealign
import nipy.algorithms.slicetiming as st
from nipy.io.api import save_image

timefuncs = st.timefuncs.SLICETIME_FUNCTIONS

__all__ = ["space_time_realign", "aff2euler"]


def aff2euler(affine):
    """
    Compute Euler angles from 4 x 4 `affine`

    Parameters
    ----------
    affine : 4 by 4 array
        An affine transformation matrix

    Returns
    -------
    The Euler angles associated with the affine
    """
    return euler.mat2euler(aff2rot_zooms(affine)[0])


def aff2rot_zooms(affine):
    """
    Compute a rotation matrix and zooms from 4 x 4 `affine`

    Parameters
    ----------
    affine : 4 by 4 array
        An affine transformation matrix

    Returns
    -------
    R: 3 by 3 array
       A rotation matrix in 3D

    zooms: length 3 1-d array
         Vector with voxel sizes.
    """
    RZS = affine[:3, :3]
    zooms = np.sqrt(np.sum(RZS * RZS, axis=0))
    RS = RZS / zooms
    # Adjust zooms to make RS correspond (below) to a true
    # rotation matrix.
    if npl.det(RS) < 0:
        zooms[0] *= -1
        RS[:,0] *= -1
    # retrieve rotation matrix from RS with polar decomposition.
    # Discard shears
    P, S, Qs = npl.svd(RS)
    R = np.dot(P, Qs)
    return R, zooms


def space_time_realign(input, tr, slice_order='descending', slice_dim=2,
                       slice_dir=1, apply=True, make_figure=False,
                       out_name=None):
    """
    This is a scripting interface to `nipy.algorithms.registration.SpaceTimeRealign` 

    Parameters
    ----------
    input : str or list
        A full path to a file-name (4D nifti time-series) , or to a directory
        containing 4D nifti time-series, or a list of full-paths to files.
    tr : float
        The repetition time
    slice_order : str (optional)
        This is the order of slice-times in the acquisition. This is used as a
        key into the ``SLICETIME_FUNCTIONS`` dictionary from
        :mod:`nipy.algorithms.slicetiming.timefuncs`. Default: 'descending'.
    slice_dim : int (optional) 
        Denotes the axis in `images` that is the slice axis.  In a 4D image,
        this will often be axis = 2 (default).
    slice_dir : int (optional)
        1 if the slices were acquired slice 0 first (default), slice -1 last,
        or -1 if acquire slice -1 first, slice 0 last.
    apply : bool (optional)
        Whether to apply the transformation and produce an output. Default:
        True. 
    make_figure : bool (optional)
        Whether to generate a .png figure with the parameters across scans.
    out_name : bool (optional)
        Specify an output location (full path) for the files that are
        generated. Default: generate files in the path of the inputs (with an
        `_mc` suffix added to the file-names.

    Returns
    -------
    transforms : ndarray
        An (n_times_points,) shaped array containing 
       `nipy.algorithms.registration.affine.Rigid` class instances for each time
        point in the time-series. These can be used as affine transforms by
        referring to their `.as_affine` attribute.
    """
    if make_figure:
        if not HAVE_MPL:
            e_s ="You need to have matplotlib installed to run this function"
            e_s += " with `make_figure` set to `True`"
            raise RuntimeError(e_s)

    # If we got only a single file, we motion correct that one:
    if op.isfile(input):
        if not (input.endswith('.nii') or input.endswith('.nii.gz')):
            e_s = "Input needs to be a nifti file ('.nii' or '.nii.gz'"
            raise ValueError(e_s)
        fnames = [input]
        input = nib.load(input)
    # If this is a full-path to a directory containing files, it's still a
    # string: 
    elif isinstance(input, str):
        list_of_files = os.listdir(input)
        fnames = [op.join(input, f) for f in np.sort(list_of_files)
                  if (f.endswith('.nii') or f.endswith('.nii.gz')) ]
        input = [nib.load(x) for x in fnames]
    # Assume that it's a list of full-paths to files:
    else:
       input = [nib.load(x) for x in input]

    slice_times = timefuncs[slice_order]
    slice_info = [slice_dim,
                  slice_dir]

    reggy = SpaceTimeRealign(input,
                             tr,
                             slice_times,
                             slice_info)

    reggy.estimate(align_runs=True)

    # We now have the transformation parameters in here:
    transforms = np.squeeze(np.array(reggy._transforms))
    rot = np.array([t.rotation for t in transforms])
    trans = np.array([t.translation for t in transforms])

    if apply:
        new_reggy = reggy.resample(align_runs=True)
        for run_idx, new_im in enumerate(new_reggy):
            # Fix output TR - it was probably lost in the image realign step
            assert new_im.affine.shape == (5, 5)
            new_im.affine[:] = new_im.affine.dot(np.diag([1, 1, 1, tr, 1]))
            # Save it out to a '.nii.gz' file:
            froot, ext, trail_ext = splitext_addext(fnames[run_idx])
            path, fname = op.split(froot)
            # We retain the file-name adding '_mc' regardless of where it's
            # saved
            new_path = path if out_name is None else out_name
            save_image(new_im, op.join(new_path, fname + '_mc.nii.gz'))

    if make_figure:
        # Delay MPL plotting import to latest moment to avoid errors trying
        # import the default MPL backend (such as tkinter, which may not be
        # installed). See: https://github.com/nipy/nipy/issues/414
        import matplotlib.pyplot as plt
        figure, ax = plt.subplots(2)
        figure.set_size_inches([8, 6])
        ax[0].plot(rot)
        ax[0].set_xlabel('Time (TR)')
        ax[0].set_ylabel('Translation (mm)')
        ax[1].plot(trans)
        ax[1].set_xlabel('Time (TR)')
        ax[1].set_ylabel('Rotation (radians)')
        figure.savefig(op.join(os.path.split(fnames[0])[0],
                                    'mc_params.png'))

    return transforms



# vi: set ft=python sts=4 ts=4 sw=4 et:
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
"""
The main routine of this module implement the LandmarkRegions class,
that is used to represent Regions of interest at the population level
(in a template space).

This has been used in
Thirion et al. Structural Analysis of fMRI
Data Revisited: Improving the Sensitivity and Reliability of fMRI
Group Studies.  IEEE TMI 2007

Author : Bertrand Thirion, 2006-2013
"""
from __future__ import print_function
from __future__ import absolute_import

#autoindent

import numpy as np
from scipy import stats


def _threshold_weight_map(x, fraction):
    """threshold a positive map in order to retain a certain fraction of the
    total value"""
    sorted_x = - np.sort(- x)
    if fraction < sorted_x[0] / x.sum():
        return np.zeros_like(x)

    idx = np.where(np.cumsum(sorted_x) < fraction * x.sum())[0][-1]
    x[x < sorted_x[idx]] = 0
    return x


class LandmarkRegions(object):
    """
    This class is intended to represent a set of inter-subject regions
    It should inherit from some abstract multiple ROI class,
    not implemented yet.
    """

    def __init__(self, domain, k, indiv_coord, subjects, confidence):
        """ Building the landmark_region

        Parameters
        ----------
        domain: ROI instance
                defines the spatial context of the SubDomains
        k: int,
           the number of landmark regions considered
        indiv_coord:  k-length list of arrays,
                      coordinates of the nodes in some embedding space.
        subjects: k-length list of integers
              these correspond to an ROI feature:
              the subject index of individual regions
        confidence: k-length list of arrays,
                    confidence values for the regions (0 is low, 1 is high)
        """
        self.domain = domain
        self.k = int(k)
        if len(indiv_coord) != k:
            raise ValueError('len(indiv_coord) should be equal to %d' % k)
        if len(subjects) != k:
            raise ValueError('len(subjects) should be equal to %d' % k)
        if len(confidence) != k:
            raise ValueError('len(confidence) should be equal to %d' % k)
        self.position = indiv_coord
        self.subjects = subjects
        self.confidence = confidence

    def centers(self):
        """returns the average of the coordinates for each region
        """
        pos = self.position
        centers_ = np.array([np.mean(pos[k], 0) for k in range(self.k)])
        return centers_

    def kernel_density(self, k=None, coord=None, sigma=1.):
        """ Compute the density of a component as a kde

        Parameters
        ----------
        k: int (<= self.k) or None
           component upon which the density is computed
           if None, the sum is taken over k
        coord: array of shape(n, self.dom.em_dim), optional
            a set of input coordinates
        sigma: float, optional
               kernel size

        Returns
        -------
        kde: array of shape(n)
             the density sampled at the coords
        """
        from nipy.algorithms.utils.fast_distance import euclidean_distance
        if coord is None:
            coord = self.domain.coord
        if k is None:
            kde = np.zeros(coord.shape[0])
            for k in range(self.k):
                pos = self.position[k]
                dist = euclidean_distance(pos, coord)
                kde += np.exp(- dist ** 2 / (2 * sigma ** 2)).sum(0)
        else:
            k = int(k)
            pos = self.position[k]
            dist = euclidean_distance(pos, coord)
            kde = np.exp(- dist ** 2 / (2 * sigma ** 2)).sum(0)
        return kde / (2 * np.pi * sigma ** 2) ** (pos.shape[1] / 2)

    def map_label(self, coord=None, pval=1., sigma=1.):
        """Sample the set of landmark regions
        on the proposed coordiante set cs, assuming a Gaussian shape

        Parameters
        ----------
        coord: array of shape(n,dim), optional,
               a set of input coordinates
        pval: float in [0,1]), optional
              cutoff for the CR, i.e.  highest posterior density threshold
        sigma: float, positive, optional
               spatial scale of the spatial model

        Returns
        -------
        label: array of shape (n): the posterior labelling
        """
        if coord is None:
            coord = self.domain.coord
        label = - np.ones(coord.shape[0])
        null_density = 1. / self.domain.local_volume.sum()
        if self.k > 0:
            aux = - np.zeros((coord.shape[0], self.k))
            for k in range(self.k):
                kde = self.kernel_density(k, coord, sigma)
                aux[:, k] = _threshold_weight_map(kde, pval)

            aux[aux < null_density] = 0
            maux = np.max(aux, 1)
            label[maux > 0] = np.argmax(aux, 1)[maux > 0]
        return label

    def show(self):
        """function to print basic information on self
        """
        centers = self.centers()
        subjects = self.subjects
        prevalence = self.roi_prevalence()
        print("index", "prevalence", "mean_position", "individuals")
        for i in range(self.k):
            print(i, prevalence[i], centers[i], np.unique(subjects[i]))

    def roi_prevalence(self):
        """ Return a confidence index over the different rois

        Returns
        -------
        confid: array of shape self.k
               the population_prevalence
        """
        prevalence_ = np.zeros(self.k)
        subjects = self.subjects
        for j in range(self.k):
            subjj = subjects[j]
            conf = self.confidence[j]
            for ls in np.unique(subjj):
                lmj = 1 - np.prod(1 - conf[subjj == ls])
                prevalence_[j] += lmj
        return prevalence_


def build_landmarks(domain, coords, subjects, labels, confidence=None,
                    prevalence_pval=0.95, prevalence_threshold=0, sigma=1.):
    """
    Given a list of hierarchical ROIs, and an associated labelling, this
    creates an Amer structure wuch groups ROIs with the same label.

    Parameters
    ----------
    domain: discrete_domain.DiscreteDomain instance,
            description of the spatial context of the landmarks
    coords: array of shape(n, 3)
            Sets of coordinates for the different objects
    subjects: array of shape (n), dtype = np.int
              indicators of the dataset the objects come from
    labels: array of shape (n), dtype = np.int
            index of the landmark the object is associated with
    confidence: array of shape (n),
                measure of the significance of the regions
    prevalence_pval: float, optional
    prevalence_threshold: float, optional,
                   (c) A label should be present in prevalence_threshold
                   subjects with a probability>prevalence_pval
                   in order to be valid
    sigma: float optional,
          regularizing constant that defines a prior on the region extent

    Returns
    -------
    LR : None or structural_bfls.LR instance
         describing a cross-subject set of ROIs. If inference yields a null
         result, LR is set to None
    newlabel: array of shape (n)
              a relabelling of the individual ROIs, similar to u,
              that discards labels that do not fulfill the condition (c)
    """
    if confidence is None:
        confidence = np.ones(labels.size)
    intrasubj = np.concatenate([np.arange(np.sum(subjects == s))
                                for s in np.unique(subjects)])

    coordinates = []
    subjs = []
    pps = []
    n_labels = int(labels.max() + 1)
    valid = np.zeros(n_labels).astype(np.int)

    # do some computation to find which regions are worth reporting
    for i in np.unique(labels[labels > - 1]):
        mean_c, var_c = 0., 0.
        subjects_i = subjects[labels == i]
        for subject_i in np.unique(subjects_i):
            confidence_i = 1 - np.prod(1 - confidence[(labels == i) *
                                                      (subjects == subject_i)])
            mean_c += confidence_i
            var_c += confidence_i * (1 - confidence_i)

        # If noise is too low the variance is 0: ill-defined:
        var_c = max(var_c, 1e-14)

        # if above threshold, get some information to create the landmarks
        if (stats.norm.sf(prevalence_threshold, mean_c, np.sqrt(var_c)) >
            prevalence_pval):
            coord = np.vstack([
                    coords[subjects == s][k] for (k, s) in zip(
                        intrasubj[labels == i],
                        subjects[labels == i])])
            valid[i] = 1
            coordinates.append(coord)
            subjs.append(subjects_i)
            pps.append(confidence[labels == i])

    # relabel the ROIs
    maplabel = - np.ones(n_labels).astype(np.int)
    maplabel[valid > 0] = np.cumsum(valid[valid > 0]) - 1

    # create the landmark regions structure
    LR = LandmarkRegions(domain, np.sum(valid), indiv_coord=coordinates,
                         subjects=subjs, confidence=pps)
    return LR, maplabel



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The Slicer classes.

The main purpose of these classes is to have auto adjust of axes size to
the data with different layout of cuts.
"""
from __future__ import absolute_import

import numpy as np

from nipy.utils import is_iterable

from nipy.utils.skip_test import skip_if_running_nose

try:
    import matplotlib as mpl
    import pylab as pl
    from matplotlib import transforms
except ImportError:
    skip_if_running_nose('Could not import matplotlib')


# Local imports
from .coord_tools import coord_transform, get_bounds, get_mask_bounds, \
        find_cut_coords
from .edge_detect import _edge_map
from . import cm
from ..datasets import VolumeImg

################################################################################
# Bugware to have transparency work OK with MPL < .99.1
if mpl.__version__ < '0.99.1':
    # We wrap the lut as a callable and replace its evalution to put
    # alpha to zero where the mask is true. This is what is done in
    # MPL >= .99.1
    from matplotlib import colors
    class CMapProxy(colors.Colormap):
        def __init__(self, lut):
            self.__lut = lut

        def __call__(self, arr, *args, **kwargs):
            results = self.__lut(arr, *args, **kwargs)
            if not isinstance(arr, np.ma.MaskedArray):
                return results
            else:
                results[arr.mask, -1] = 0
            return results

        def __getattr__(self, attr):
            # Dark magic: we are delegating any call to the lut instance
            # we wrap
            return self.__dict__.get(attr, getattr(self.__lut, attr))


def _xyz_order(map, affine):
    img = VolumeImg(map, affine=affine, world_space='mine')
    img = img.xyz_ordered(resample=True, copy=False)
    map = img.get_data()
    affine = img.affine
    return map, affine


################################################################################
# class CutAxes
################################################################################

class CutAxes(object):
    """ An MPL axis-like object that displays a cut of 3D volumes
    """

    def __init__(self, ax, direction, coord):
        """ An MPL axis-like object that displays a cut of 3D volumes

            Parameters
            ==========
            ax: a MPL axes instance
                The axes in which the plots will be drawn
            direction: {'x', 'y', 'z'}
                The directions of the cut
            coord: float
                The coordinnate along the direction of the cut
        """
        self.ax = ax
        self.direction = direction
        self.coord = coord
        self._object_bounds = list()


    def do_cut(self, map, affine):
        """ Cut the 3D volume into a 2D slice

            Parameters
            ==========
            map: 3D ndarray
                The 3D volume to cut
            affine: 4x4 ndarray
                The affine of the volume
        """
        coords = [0, 0, 0]
        coords['xyz'.index(self.direction)] = self.coord
        x_map, y_map, z_map = [int(np.round(c)) for c in
                               coord_transform(coords[0],
                                               coords[1],
                                               coords[2],
                                               np.linalg.inv(affine))]
        if self.direction == 'y':
            cut = np.rot90(map[:, y_map, :])
        elif self.direction == 'x':
            cut = np.rot90(map[x_map, :, :])
        elif self.direction == 'z':
            cut = np.rot90(map[:, :, z_map])
        else:
            raise ValueError('Invalid value for direction %s' %
                             self.direction)
        return cut


    def draw_cut(self, cut, data_bounds, bounding_box,
                  type='imshow', **kwargs):
        # kwargs massaging
        kwargs['origin'] = 'upper'
        if mpl.__version__ < '0.99.1':
            cmap = kwargs.get('cmap',
                        pl.cm.cmap_d[pl.rcParams['image.cmap']])
            kwargs['cmap'] = CMapProxy(cmap)

        if self.direction == 'y':
            (xmin, xmax), (_, _), (zmin, zmax) = data_bounds
            (xmin_, xmax_), (_, _), (zmin_, zmax_) = bounding_box
        elif self.direction == 'x':
            (_, _), (xmin, xmax), (zmin, zmax) = data_bounds
            (_, _), (xmin_, xmax_), (zmin_, zmax_) = bounding_box
        elif self.direction == 'z':
            (xmin, xmax), (zmin, zmax), (_, _) = data_bounds
            (xmin_, xmax_), (zmin_, zmax_), (_, _) = bounding_box
        else:
            raise ValueError('Invalid value for direction %s' %
                             self.direction)
        ax = self.ax
        getattr(ax, type)(cut, extent=(xmin, xmax, zmin, zmax), **kwargs)

        self._object_bounds.append((xmin_, xmax_, zmin_, zmax_))
        ax.axis(self.get_object_bounds())


    def get_object_bounds(self):
        """ Return the bounds of the objects on this axes.
        """
        if len(self._object_bounds) == 0:
            # Nothing plotted yet
            return -.01, .01, -.01, .01
        xmins, xmaxs, ymins, ymaxs = np.array(self._object_bounds).T
        xmax = max(xmaxs.max(), xmins.max())
        xmin = min(xmins.min(), xmaxs.min())
        ymax = max(ymaxs.max(), ymins.max())
        ymin = min(ymins.min(), ymaxs.min())
        return xmin, xmax, ymin, ymax


    def draw_left_right(self, size, bg_color, **kwargs):
        if self.direction == 'x':
            return
        ax = self.ax
        ax.text(.1, .95, 'L',
                transform=ax.transAxes,
                horizontalalignment='left',
                verticalalignment='top',
                size=size,
                bbox=dict(boxstyle="square,pad=0",
                            ec=bg_color, fc=bg_color, alpha=1),
                **kwargs)

        ax.text(.9, .95, 'R',
                transform=ax.transAxes,
                horizontalalignment='right',
                verticalalignment='top',
                size=size,
                bbox=dict(boxstyle="square,pad=0",
                            ec=bg_color, fc=bg_color, alpha=1),
                **kwargs)


    def draw_position(self, size, bg_color, **kwargs):
        ax = self.ax
        ax.text(0, 0, '%s=%i' % (self.direction, self.coord),
                transform=ax.transAxes,
                horizontalalignment='left',
                verticalalignment='bottom',
                size=size,
                bbox=dict(boxstyle="square,pad=0",
                            ec=bg_color, fc=bg_color, alpha=1),
                **kwargs)


################################################################################
# class BaseSlicer
################################################################################

class BaseSlicer(object):
    """ The main purpose of these class is to have auto adjust of axes size
        to the data with different layout of cuts.
    """
    # This actually encodes the figsize for only one axe
    _default_figsize = [2.2, 2.6]

    def __init__(self, cut_coords, axes=None, black_bg=False):
        """ Create 3 linked axes for plotting orthogonal cuts.

            Parameters
            ----------
            cut_coords: 3 tuple of ints
                The cut position, in world space.
            axes: matplotlib axes object, optional
                The axes that will be subdivided in 3.
            black_bg: boolean, optional
                If True, the background of the figure will be put to
                black. If you whish to save figures with a black background, 
                you will need to pass "facecolor='k', edgecolor='k'" to 
                pylab's savefig.

        """
        self._cut_coords = cut_coords
        if axes is None:
            axes = pl.axes((0., 0., 1., 1.))
            axes.axis('off')
        self.frame_axes = axes
        axes.set_zorder(1)
        bb = axes.get_position()
        self.rect = (bb.x0, bb.y0, bb.x1, bb.y1)
        self._black_bg = black_bg
        self._init_axes()


    @staticmethod
    def find_cut_coords(data=None, affine=None, threshold=None,
                        cut_coords=None):
        # Implement this as a staticmethod or a classmethod when
        # subclassing
        raise NotImplementedError

    @classmethod
    def init_with_figure(cls, data=None, affine=None, threshold=None,
                         cut_coords=None, figure=None, axes=None,
                         black_bg=False, leave_space=False):
        cut_coords = cls.find_cut_coords(data, affine, threshold,
                                         cut_coords)
        if isinstance(axes, pl.Axes) and figure is None:
            figure = axes.figure

        if not isinstance(figure, pl.Figure):
            # Make sure that we have a figure
            figsize = cls._default_figsize[:]
            # Adjust for the number of axes
            figsize[0] *= len(cut_coords)
            facecolor = 'k' if black_bg else 'w'

            if leave_space:
                figsize[0] += 3.4
            figure = pl.figure(figure, figsize=figsize,
                            facecolor=facecolor)
        else:
            if isinstance(axes, pl.Axes):
                assert axes.figure is figure, ("The axes passed are not "
                "in the figure")

        if axes is None:
            axes = [0., 0., 1., 1.]
            if leave_space:
                axes = [0.3, 0, .7, 1.]
        if is_iterable(axes):
            axes = figure.add_axes(axes)
        # People forget to turn their axis off, or to set the zorder, and
        # then they cannot see their slicer
        axes.axis('off')
        return cls(cut_coords, axes, black_bg)


    def title(self, text, x=0.01, y=0.99, size=15, color=None,
                bgcolor=None, alpha=1, **kwargs):
        """ Write a title to the view.

            Parameters
            ----------
            text: string
                The text of the title
            x: float, optional
                The horizontal position of the title on the frame in 
                fraction of the frame width.
            y: float, optional
                The vertical position of the title on the frame in 
                fraction of the frame height.
            size: integer, optional
                The size of the title text.
            color: matplotlib color specifier, optional
                The color of the font of the title.
            bgcolor: matplotlib color specifier, optional
                The color of the background of the title.
            alpha: float, optional
                The alpha value for the background.
            kwargs:
                Extra keyword arguments are passed to matplotlib's text
                function.
        """
        if color is None:
            color = 'k' if self._black_bg else 'w'
        if bgcolor is None:
            bgcolor = 'w' if self._black_bg else 'k'
        self.frame_axes.text(x, y, text, 
                    transform=self.frame_axes.transAxes,
                    horizontalalignment='left',
                    verticalalignment='top',
                    size=size, color=color,
                    bbox=dict(boxstyle="square,pad=.3", 
                              ec=bgcolor, fc=bgcolor, alpha=alpha),
                    **kwargs)


    def plot_map(self, map, affine, threshold=None, **kwargs):
        """ Plot a 3D map in all the views.

            Parameters
            -----------
            map: 3D ndarray
                The 3D map to be plotted. If it is a masked array, only
                the non-masked part will be plotted.
            affine: 4x4 ndarray
                The affine matrix giving the transformation from voxel
                indices to world space.
            threshold : a number, None, or 'auto'
                If None is given, the maps are not thresholded.
                If a number is given, it is used to threshold the maps:
                values below the threshold are plotted as transparent.
            kwargs:
                Extra keyword arguments are passed to imshow.
        """
        if threshold is not None:
            if threshold == 0:
                map = np.ma.masked_equal(map, 0, copy=False)
            else:
                map = np.ma.masked_inside(map, -threshold, threshold, 
                                          copy=False)

        self._map_show(map, affine, type='imshow', **kwargs)


    def contour_map(self, map, affine, **kwargs):
        """ Contour a 3D map in all the views.

            Parameters
            -----------
            map: 3D ndarray
                The 3D map to be plotted. If it is a masked array, only
                the non-masked part will be plotted.
            affine: 4x4 ndarray
                The affine matrix giving the transformation from voxel
                indices to world space.
            kwargs:
                Extra keyword arguments are passed to contour.
        """
        self._map_show(map, affine, type='contour', **kwargs)


    def _map_show(self, map, affine, type='imshow', **kwargs):
        map, affine = _xyz_order(map, affine)

        data_bounds = get_bounds(map.shape, affine)
        (xmin, xmax), (ymin, ymax), (zmin, zmax) = data_bounds

        xmin_, xmax_, ymin_, ymax_, zmin_, zmax_ = \
                                        xmin, xmax, ymin, ymax, zmin, zmax
        if hasattr(map, 'mask'):
            not_mask = np.logical_not(map.mask)
            xmin_, xmax_, ymin_, ymax_, zmin_, zmax_ = \
                            get_mask_bounds(not_mask, affine)
            if kwargs.get('vmin') is None and kwargs.get('vmax') is None:
                # Avoid dealing with masked arrays: they are slow
                if not np.any(not_mask):
                    # Everything is masked
                    vmin = vmax = 0
                else:
                    masked_map = np.asarray(map)[not_mask]
                    vmin = masked_map.min()
                    vmax = masked_map.max()
                if kwargs.get('vmin') is None:
                    kwargs['vmin'] = vmin
                if kwargs.get('max') is None:
                    kwargs['vmax'] = vmax
        else:
            if not 'vmin' in kwargs:
                kwargs['vmin'] = map.min()
            if not 'vmax' in kwargs:
                kwargs['vmax'] = map.max()

        bounding_box = (xmin_, xmax_), (ymin_, ymax_), (zmin_, zmax_)

        # For each ax, cut the data and plot it
        for cut_ax in self.axes.values():
            try:
                cut = cut_ax.do_cut(map, affine)
            except IndexError:
                # We are cutting outside the indices of the data
                continue
            cut_ax.draw_cut(cut, data_bounds, bounding_box,
                            type=type, **kwargs)


    def edge_map(self, map, affine, color='r'):
        """ Plot the edges of a 3D map in all the views.

            Parameters
            -----------
            map: 3D ndarray
                The 3D map to be plotted. If it is a masked array, only
                the non-masked part will be plotted.
            affine: 4x4 ndarray
                The affine matrix giving the transformation from voxel
                indices to world space.
            color: matplotlib color: string or (r, g, b) value
                The color used to display the edge map
        """
        map, affine = _xyz_order(map, affine)
        kwargs = dict(cmap=cm.alpha_cmap(color=color))
        data_bounds = get_bounds(map.shape, affine)

        # For each ax, cut the data and plot it
        for cut_ax in self.axes.values():
            try:
                cut = cut_ax.do_cut(map, affine)
                edge_mask = _edge_map(cut)
            except IndexError:
                # We are cutting outside the indices of the data
                continue
            cut_ax.draw_cut(edge_mask, data_bounds, data_bounds,
                            type='imshow', **kwargs)

    def annotate(self, left_right=True, positions=True, size=12, **kwargs):
        """ Add annotations to the plot.

            Parameters
            ----------
            left_right: boolean, optional
                If left_right is True, annotations indicating which side
                is left and which side is right are drawn.
            positions: boolean, optional
                If positions is True, annotations indicating the
                positions of the cuts are drawn.
            size: integer, optional
                The size of the text used.
            kwargs:
                Extra keyword arguments are passed to matplotlib's text
                function.
        """
        kwargs = kwargs.copy()
        if not 'color' in kwargs:
            if self._black_bg:
                kwargs['color'] = 'w'
            else:
                kwargs['color'] = 'k'

        bg_color = ('k' if self._black_bg else 'w')
        if left_right:
            for cut_ax in self.axes.values():
                cut_ax.draw_left_right(size=size, bg_color=bg_color,
                                       **kwargs)

        if positions:
            for cut_ax in self.axes.values():
                cut_ax.draw_position(size=size, bg_color=bg_color,
                                       **kwargs)


################################################################################
# class OrthoSlicer
################################################################################

class OrthoSlicer(BaseSlicer):
    """ A class to create 3 linked axes for plotting orthogonal
        cuts of 3D maps.

        Attributes
        ----------

        axes: dictionnary of axes
            The 3 axes used to plot each view.
        frame_axes: axes
            The axes framing the whole set of views.

        Notes
        -----

        The extent of the different axes are adjusted to fit the data
        best in the viewing area.
    """
    @staticmethod
    def find_cut_coords(data=None, affine=None, threshold=None,
                        cut_coords=None):
        if cut_coords is None:
            if data is None or data is False:
                cut_coords = (0, 0, 0)
            else:
                x_map, y_map, z_map = find_cut_coords(data,
                                        activation_threshold=threshold)
                cut_coords = coord_transform(x_map, y_map, z_map, affine)
        return cut_coords


    def _init_axes(self):
        x0, y0, x1, y1 = self.rect
        # Create our axes:
        self.axes = dict()
        for index, direction in enumerate(('y', 'x', 'z')):
            ax = pl.axes([0.3*index*(x1-x0) + x0, y0, .3*(x1-x0), y1-y0])
            ax.axis('off')
            coord = self._cut_coords['xyz'.index(direction)]
            cut_ax = CutAxes(ax, direction, coord)
            self.axes[direction] = cut_ax
            ax.set_axes_locator(self._locator)


    def _locator(self, axes, renderer):
        """ The locator function used by matplotlib to position axes.
            Here we put the logic used to adjust the size of the axes.
        """
        x0, y0, x1, y1 = self.rect
        width_dict = dict()
        cut_ax_dict = self.axes
        x_ax = cut_ax_dict['x']
        y_ax = cut_ax_dict['y']
        z_ax = cut_ax_dict['z']
        for cut_ax in cut_ax_dict.values():
            bounds = cut_ax.get_object_bounds()
            if not bounds:
                # This happens if the call to _map_show was not
                # succesful. As it happens asyncroniously (during a
                # refresh of the figure) we capture the problem and
                # ignore it: it only adds a non informative traceback
                bounds = [0, 1, 0, 1]
            xmin, xmax, ymin, ymax = bounds
            width_dict[cut_ax.ax] = (xmax - xmin)
        total_width = float(sum(width_dict.values()))
        for ax, width in width_dict.items():
            width_dict[ax] = width/total_width*(x1 -x0)
        left_dict = dict()
        left_dict[y_ax.ax] = x0
        left_dict[x_ax.ax] = x0 + width_dict[y_ax.ax]
        left_dict[z_ax.ax] = x0 + width_dict[x_ax.ax] + width_dict[y_ax.ax]
        return transforms.Bbox([[left_dict[axes], y0],
                          [left_dict[axes] + width_dict[axes], y1]])


    def draw_cross(self, cut_coords=None, **kwargs):
        """ Draw a crossbar on the plot to show where the cut is
            performed.

            Parameters
            ----------
            cut_coords: 3-tuple of floats, optional
                The position of the cross to draw. If none is passed, the
                ortho_slicer's cut coordinnates are used.
            kwargs:
                Extra keyword arguments are passed to axhline
        """
        if cut_coords is None:
            cut_coords = self._cut_coords
        x, y, z = cut_coords
        kwargs = kwargs.copy()
        if not 'color' in kwargs:
            if self._black_bg:
                kwargs['color'] = '.8'
            else:
                kwargs['color'] = 'k'
        ax = self.axes['y'].ax
        ax.axvline(x, ymin=.05, ymax=.95, **kwargs)
        ax.axhline(z, **kwargs)

        ax = self.axes['x'].ax
        ax.axvline(y, ymin=.05, ymax=.95, **kwargs)
        ax.axhline(z, xmax=.95, **kwargs)

        ax = self.axes['z'].ax
        ax.axvline(x, ymin=.05, ymax=.95, **kwargs)
        ax.axhline(y, **kwargs)



def demo_ortho_slicer():
    """ A small demo of the OrthoSlicer functionality.
    """
    pl.clf()
    oslicer = OrthoSlicer(cut_coords=(0, 0, 0))
    from .anat_cache import _AnatCache
    map, affine, _ = _AnatCache.get_anat()
    oslicer.plot_map(map, affine, cmap=pl.cm.gray)
    return oslicer


################################################################################
# class BaseStackedSlicer
################################################################################

class BaseStackedSlicer(BaseSlicer):
    """ A class to create linked axes for plotting stacked
        cuts of 3D maps.

        Attributes
        ----------

        axes: dictionnary of axes
            The axes used to plot each view.
        frame_axes: axes
            The axes framing the whole set of views.

        Notes
        -----

        The extent of the different axes are adjusted to fit the data
        best in the viewing area.
    """
    @classmethod
    def find_cut_coords(cls, data=None, affine=None, threshold=None,
                        cut_coords=None):
        if cut_coords is None:
            if data is None or data is False:
                bounds = ((-40, 40), (-30, 30), (-30, 75))
            else:
                if hasattr(data, 'mask'):
                    mask = np.logical_not(data.mask)
                else:
                    # The mask will be anything that is fairly different
                    # from the values in the corners
                    edge_value = float(data[0, 0, 0] + data[0, -1, 0]
                                     + data[-1, 0, 0] + data[0, 0, -1]
                                     + data[-1, -1, 0] + data[-1, 0, -1]
                                     + data[0, -1, -1] + data[-1, -1, -1]
                                    )
                    edge_value /= 6
                    mask = np.abs(data - edge_value) > .005*data.ptp()
                xmin, xmax, ymin, ymax, zmin, zmax = \
                                get_mask_bounds(mask, affine)
                bounds = (xmin, xmax), (ymin, ymax), (zmin, zmax)
            lower, upper = bounds['xyz'.index(cls._direction)]
            cut_coords = np.linspace(lower, upper, 10).tolist()
        return cut_coords


    def _init_axes(self):
        x0, y0, x1, y1 = self.rect
        # Create our axes:
        self.axes = dict()
        fraction = 1./len(self._cut_coords)
        for index, coord in enumerate(self._cut_coords):
            coord = float(coord)
            ax = pl.axes([fraction*index*(x1-x0) + x0, y0,
                          fraction*(x1-x0), y1-y0])
            ax.axis('off')
            cut_ax = CutAxes(ax, self._direction, coord)
            self.axes[coord] = cut_ax
            ax.set_axes_locator(self._locator)


    def _locator(self, axes, renderer):
        """ The locator function used by matplotlib to position axes.
            Here we put the logic used to adjust the size of the axes.
        """
        x0, y0, x1, y1 = self.rect
        width_dict = dict()
        cut_ax_dict = self.axes
        for cut_ax in cut_ax_dict.values():
            bounds = cut_ax.get_object_bounds()
            if not bounds:
                # This happens if the call to _map_show was not
                # succesful. As it happens asyncroniously (during a
                # refresh of the figure) we capture the problem and
                # ignore it: it only adds a non informative traceback
                bounds = [0, 1, 0, 1]
            xmin, xmax, ymin, ymax = bounds
            width_dict[cut_ax.ax] = (xmax - xmin)
        total_width = float(sum(width_dict.values()))
        for ax, width in width_dict.items():
            width_dict[ax] = width/total_width*(x1 -x0)
        left_dict = dict()
        left = float(x0)
        for coord, cut_ax in sorted(cut_ax_dict.items()):
            left_dict[cut_ax.ax] = left
            this_width = width_dict[cut_ax.ax]
            left += this_width
        return transforms.Bbox([[left_dict[axes], y0],
                                [left_dict[axes] + width_dict[axes], y1]])


    def draw_cross(self, cut_coords=None, **kwargs):
        """ Draw a crossbar on the plot to show where the cut is
            performed.

            Parameters
            ----------
            cut_coords: 3-tuple of floats, optional
                The position of the cross to draw. If none is passed, the
                ortho_slicer's cut coordinnates are used.
            kwargs:
                Extra keyword arguments are passed to axhline
        """
        return


class XSlicer(BaseStackedSlicer):
    _direction = 'x'
    _default_figsize = [2.2, 2.3]


class YSlicer(BaseStackedSlicer):
    _direction = 'y'
    _default_figsize = [2.6, 2.3]


class ZSlicer(BaseStackedSlicer):
    _direction = 'z'


SLICERS = dict(ortho=OrthoSlicer,
               x=XSlicer,
               y=YSlicer,
               z=ZSlicer)



""" Testing realfuncs module
"""

from os.path import dirname, join as pjoin
from itertools import product

import numpy as np

from ..realfuncs import dct_ii_basis, dct_ii_cut_basis

from nose.tools import assert_raises

from numpy.testing import (assert_almost_equal,
                           assert_array_equal)


HERE = dirname(__file__)


def test_dct_ii_basis():
    # Test DCT-II basis
    for N in (5, 10, 100):
        spm_fname = pjoin(HERE, 'dct_{0}.txt'.format(N))
        spm_mtx = np.loadtxt(spm_fname)
        vol_times = np.arange(N) * 15. + 3.2
        our_dct = dct_ii_basis(vol_times)
        # Check dot products of columns
        sq_col_lengths = np.ones(N) * N / 2.
        sq_col_lengths[0] = N
        assert_almost_equal(our_dct.T.dot(our_dct),
                            np.diag(sq_col_lengths))
        col_lengths = np.sqrt(sq_col_lengths)
        assert_almost_equal(our_dct / col_lengths, spm_mtx)
        # Normalize length
        our_normed_dct = dct_ii_basis(vol_times, normcols=True)
        assert_almost_equal(our_normed_dct, spm_mtx)
        assert_almost_equal(our_normed_dct.T.dot(our_normed_dct), np.eye(N))
        for i in range(N):
            assert_almost_equal(dct_ii_basis(vol_times, i) / col_lengths[:i],
                                spm_mtx[:, :i])
            assert_almost_equal(dct_ii_basis(vol_times, i, True),
                                spm_mtx[:, :i])
    vol_times[0] += 0.1
    assert_raises(ValueError, dct_ii_basis, vol_times)


def test_dct_ii_cut_basis():
    # DCT-II basis with cut frequency
    for dt, cut_period, N in product((0.1, 1.1),
                                     (10.1, 20.1),
                                     (20, 100, 1000)):
        times = np.arange(N) * dt
        order = int(np.floor(2 * N * 1./ cut_period * dt))
        dct_vals = dct_ii_cut_basis(times, cut_period)
        if order == 0:
            assert_array_equal(dct_vals, np.ones((N, 1)))
            continue
        dct_expected = np.ones((N, order))
        dct_expected[:, :-1] = dct_ii_basis(times, order, normcols=True)[:, 1:]
        assert_array_equal(dct_vals, dct_expected)



"""
Generic implementation of multiple regression analysis under noisy
measurements.
"""
from __future__ import absolute_import

import numpy as np

nonzero = lambda x: np.maximum(x, 1e-25)


def two_level_glm(y, vy, X, niter=10):
    """
    Inference of a mixed-effect linear model using the variational
    Bayes algorithm.

    Parameters
    ----------
    y : array-like
      Array of observations. Shape should be (n, ...) where n is the
      number of independent observations per unit.

    vy : array-like
      First-level variances associated with the observations. Should
      be of the same shape as Y.

    X : array-like
      Second-level design matrix. Shape should be (n, p) where n is
      the number of observations per unit, and p is the number of
      regressors.

    Returns
    -------
    beta : array-like
      Effect estimates (posterior means)

    s2 : array-like
      Variance estimates. The posterior variance matrix of beta[:, i]
      may be computed by s2[:, i] * inv(X.T * X)

    dof : float
      Degrees of freedom as per the variational Bayes approximation
      (simply, the number of observations minus the number of
      independent regressors)
    """
    # Number of observations, regressors and points
    nobs = X.shape[0]
    if X.ndim == 1:
        nreg = 1
    else:
        nreg = X.shape[1]
    if nobs <= nreg:
        raise ValueError('Too many regressors compared to data size')
    if y.ndim == 1:
        npts = 1
    else:
        npts = np.prod(y.shape[1:])

    # Reshape input arrays
    X = X.reshape((nobs, nreg))
    y = np.reshape(y, (nobs, npts))
    vy = nonzero(np.reshape(vy, (nobs, npts)))

    # Degrees of freedom
    dof = float(nobs - nreg)

    # Compute the pseudo-inverse matrix
    pinvX = np.linalg.pinv(X)

    # Initialize outputs
    b = np.zeros((nreg, npts))
    zfit = np.zeros((nobs, npts))
    s2 = np.inf

    # VB loop
    for it in range(niter):

        # Update distribution of "true" effects
        w1 = 1 / vy
        w2 = 1 / nonzero(s2)
        vz = 1 / (w1 + w2)
        z = vz * (w1 * y + w2 * zfit)

        # Update distribution of population parameters
        b = np.dot(pinvX, z)
        zfit = np.dot(X, b)
        s2 = np.sum((z - zfit) ** 2 + vz, 0) / dof

    # Ouput arrays
    B = np.reshape(b, [nreg] + list(y.shape[1:]))
    S2 = np.reshape(s2, list(y.shape[1:]))

    return B, S2, dof



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Misc tools to find activations and cut on maps
"""
from __future__ import absolute_import

# Author: Gael Varoquaux <gael dot varoquaux at normalesup dot org>
# License: BSD

import warnings

# Standard scientific libraries imports (more specific imports are
# delayed, so that the part module can be used without them).
import numpy as np
from scipy import stats, ndimage

# Local imports
from ..mask import largest_cc
from ..datasets.transforms.affine_utils import get_bounds


################################################################################
# Functions for automatic choice of cuts coordinates
################################################################################

def coord_transform(x, y, z, affine):
    """ Convert x, y, z coordinates from one image space to another space.

    Warning: x, y and z have Talairach ordering, not 3D numpy image ordering.

    Parameters
    ----------
    x : number or ndarray
        The x coordinates in the input space
    y : number or ndarray
        The y coordinates in the input space
    z : number or ndarray
        The z coordinates in the input space
    affine : 2D 4x4 ndarray
        affine that maps from input to output space.

    Returns
    -------
    x : number or ndarray
        The x coordinates in the output space
    y : number or ndarray
        The y coordinates in the output space
    z : number or ndarray
        The z coordinates in the output space
    """
    coords = np.c_[np.atleast_1d(x).flat, 
                   np.atleast_1d(y).flat, 
                   np.atleast_1d(z).flat,
                   np.ones_like(np.atleast_1d(z).flat)].T
    x, y, z, _ = np.dot(affine, coords)
    return x.squeeze(), y.squeeze(), z.squeeze()


def find_cut_coords(map, mask=None, activation_threshold=None):
    """ Find the center of the largest activation connect component.

        Parameters
        -----------
        map : 3D ndarray
            The activation map, as a 3D image.
        mask : 3D ndarray, boolean, optional
            An optional brain mask.
        activation_threshold : float, optional
            The lower threshold to the positive activation. If None, the 
            activation threshold is computed using find_activation.

        Returns
        -------
        x: float
            the x coordinate in voxels.
        y: float
            the y coordinate in voxels.
        z: float
            the z coordinate in voxels.
    """
    # To speed up computations, we work with partial views of the array,
    # and keep track of the offset
    offset = np.zeros(3) 
    # Deal with masked arrays:
    if hasattr(map, 'mask'):
        not_mask = np.logical_not(map.mask)
        if mask is None:
            mask = not_mask
        else:
            mask *= not_mask
        map = np.asarray(map)
    my_map = map.copy()
    if mask is not None:
        slice_x, slice_y, slice_z = ndimage.find_objects(mask)[0]
        my_map = my_map[slice_x, slice_y, slice_z]
        mask = mask[slice_x, slice_y, slice_z]
        my_map *= mask
        offset += [slice_x.start, slice_y.start, slice_z.start]
    # Testing min and max is faster than np.all(my_map == 0)
    if (my_map.max() == 0) and (my_map.min() == 0):
        return .5*np.array(map.shape)
    if activation_threshold is None:
        activation_threshold = stats.scoreatpercentile(
                                    np.abs(my_map[my_map !=0]).ravel(), 80)
    mask = np.abs(my_map) > activation_threshold-1.e-15
    mask = largest_cc(mask)
    slice_x, slice_y, slice_z = ndimage.find_objects(mask)[0]
    my_map = my_map[slice_x, slice_y, slice_z]
    mask = mask[slice_x, slice_y, slice_z]
    my_map *= mask
    offset += [slice_x.start, slice_y.start, slice_z.start]
    # For the second threshold, we use a mean, as it is much faster,
    # althought it is less robust
    second_threshold = np.abs(np.mean(my_map[mask]))
    second_mask = (np.abs(my_map)>second_threshold)
    if second_mask.sum() > 50:
        my_map *= largest_cc(second_mask)
    cut_coords = ndimage.center_of_mass(np.abs(my_map))
    return cut_coords + offset


################################################################################

def get_mask_bounds(mask, affine):
    """ Return the world-space bounds occupied by a mask given an affine.

        Notes
        -----

        The mask should have only one connect component.

        The affine should be diagonal or diagonal-permuted.
    """
    (xmin, xmax), (ymin, ymax), (zmin, zmax) = get_bounds(mask.shape, affine)
    slices = ndimage.find_objects(mask)
    if len(slices) == 0:
        warnings.warn("empty mask", stacklevel=2)
    else:
        x_slice, y_slice, z_slice = slices[0]
        x_width, y_width, z_width = mask.shape
        xmin, xmax = (xmin + x_slice.start*(xmax - xmin)/x_width,
                    xmin + x_slice.stop *(xmax - xmin)/x_width)
        ymin, ymax = (ymin + y_slice.start*(ymax - ymin)/y_width,
                    ymin + y_slice.stop *(ymax - ymin)/y_width)
        zmin, zmax = (zmin + z_slice.start*(zmax - zmin)/z_width,
                    zmin + z_slice.stop *(zmax - zmin)/z_width)

    return xmin, xmax, ymin, ymax, zmin, zmax


def _maximally_separated_subset(x, k):
    """
    Given a set of n points x = {x_1, x_2, ..., x_n} and a positive integer
    k < n, this function returns a subset of k points which are maximally
    spaced.

    Returns
    -------
    msssk: 1D array of k floats
        computed maximally-separated subset of k elements from x

    """

    # base cases
    if k < 1: raise ValueError("k = %i < 1 is senseless." % k)
    if k == 1: return [x[len(x) // 2]]

    # would-be maximally separated subset of k (not showing the terminal nodes)
    msss = list(range(1, len(x) - 1))

    # sorting is necessary for the heuristic to work
    x = np.sort(x)

    # iteratively delete points x_j of msss, for which x_(j + 1) - x_(j - 1) is
    # smallest, untill only k - 2 points survive
    while len(msss) + 2 > k:
        # survivors
        y = np.array([x[0]] + list(x[msss]) + [x[-1]])

        # remove most troublesome point
        msss = np.delete(msss, np.argmin(y[2:] - y[:-2]))

    # return maximally separated subset of k elements
    return x[[0] + list(msss) + [len(x) - 1]]


def find_maxsep_cut_coords(map3d, affine, slicer='z', n_cuts=None,
                           threshold=None):
    """ Heuristic finds `n_cuts` with max separation along a given axis

    Parameters
    ----------
    map3d : 3D array
        the data under consideration
    affine : array shape (4, 4)
        Affine mapping between array coordinates of `map3d` and real-world
        coordinates.
    slicer : string, optional
        sectional slicer; possible values are "x", "y", or "z"
    n_cuts : None or int >= 1, optional
        Number of cuts in the plot; if None, then a default value of 5 is
        forced.
    threshold : None or float, optional
        Thresholding to be applied to the map.  Values less than `threshold`
        set to 0.  If None, no thresholding applied.

    Returns
    -------
    cuts : 1D array of length `n_cuts`
        the computed cuts

    Raises
    ------
    ValueError:
        If `slicer` not in 'xyz'
    ValueError
        If `ncuts` < 1
    """
    if n_cuts is None:
        n_cuts = 5
    if n_cuts < 1:
        raise ValueError("n_cuts = %i < 1 is senseless." % n_cuts)
    # sanitize slicer
    if slicer not in 'xyz':
        raise ValueError(
            "slicer must be one of 'x', 'y', and 'z', got '%s'." % slicer)
    slicer = "xyz".index(slicer)

    # load data
    if map3d.ndim != 3:
        raise TypeError(
            "map3d must be 3D array, got shape %iD" % map3d.ndim)

    _map3d = np.rollaxis(map3d.copy(), slicer, start=3)
    _map3d = np.abs(_map3d)
    if threshold is not None:
        _map3d[_map3d < threshold] = 0

    # count activated voxels per plane
    n_activated_voxels_per_plane = np.array([(_map3d[..., z] > 0).sum()
                                    for z in range(_map3d.shape[-1])])
    perm = np.argsort(n_activated_voxels_per_plane)
    n_activated_voxels_per_plane = n_activated_voxels_per_plane[perm]
    good_planes = np.nonzero(n_activated_voxels_per_plane > 0)[0]
    good_planes = perm[::-1][:n_cuts * 4 if n_cuts > 1 else 1]

    # cast into coord space
    good_planes = np.array([
            # map cut coord into native space
            np.dot(affine,
                   np.array([0, 0, 0, 1]  # origin
                            ) + coord * np.eye(4)[slicer])[slicer]
            for coord in good_planes])

    # compute cut_coords maximally-separated planes
    return _maximally_separated_subset(good_planes, n_cuts)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
''' Diagnostic 4d image screen '''
from __future__ import absolute_import
from os.path import join as pjoin

import warnings

import numpy as np

from ...core.api import Image, drop_io_dim
from ...core.reference.coordinate_map import input_axis_index, AxisError
from ...io.api import save_image
from ..utils import pca
from .timediff import time_slice_diffs
from .tsdiffplot import plot_tsdiffs


def screen(img4d, ncomp=10, time_axis='t', slice_axis=None):
    ''' Diagnostic screen for 4d FMRI image

    Includes PCA, tsdiffana and mean, std, min, max images.

    Parameters
    ----------
    img4d : ``Image``
       4d image file
    ncomp : int, optional
       number of component images to return.  Default is 10
    time_axis : str or int, optional
        Axis over which to do PCA, time difference analysis. Defaults to `t`
    slice_axis : None or str or int, optional
        Name or index of input axis over which to do slice analysis for time
        difference analysis.  If None, look for input axis ``slice``.  At the
        moment we then assume slice is the last non-time axis, but this last
        guess we will remove in future versions of nipy. The default will then
        be 'slice' and you'll get an error if there is no axis named 'slice'.

    Returns
    -------
    screen : dict
       with keys:

       * mean : mean image (all summaries are over last dimension)
       * std : standard deviation image
       * max : image of max
       * min : min
       * pca : 4D image of PCA component images
       * pca_res : dict of results from PCA
       * ts_res : dict of results from tsdiffana

    Examples
    --------
    >>> import nipy as ni
    >>> from nipy.testing import funcfile
    >>> img = ni.load_image(funcfile)
    >>> screen_res = screen(img)
    >>> screen_res['mean'].ndim
    3
    >>> screen_res['pca'].ndim
    4
    '''
    if img4d.ndim != 4:
        raise ValueError('Expecting a 4d image')
    data = img4d.get_data()
    cmap = img4d.coordmap
    # Get numerical index for time axis in data array
    time_axis = input_axis_index(cmap, time_axis)
    # Get numerical index for slice axis in data array
    if slice_axis is None:
        try:
            slice_axis = input_axis_index(cmap, 'slice')
        except AxisError:
            warnings.warn(
                'Future versions of nipy will not guess the slice axis '
                'from position, but only from axis name == "slice"; '
                'Please specify the slice axis by name or index to avoid '
                'this warning',
                FutureWarning,
                stacklevel=2)
            slice_axis = 2 if time_axis == 3 else 3
    else:
        slice_axis = input_axis_index(cmap, slice_axis)
    # 3D coordinate map for summary images
    cmap_3d = drop_io_dim(cmap, 't')
    screen_res = {}
    # standard processed images
    screen_res['mean'] = Image(np.mean(data, axis=time_axis), cmap_3d)
    screen_res['std'] = Image(np.std(data, axis=time_axis), cmap_3d)
    screen_res['max'] = Image(np.max(data, axis=time_axis), cmap_3d)
    screen_res['min'] = Image(np.min(data, axis=time_axis), cmap_3d)
    # PCA
    screen_res['pca_res'] = pca.pca_image(img4d,
                                          axis=time_axis,
                                          standardize=False,
                                          ncomp=ncomp)
    screen_res['pca'] = screen_res['pca_res']['basis_projections']
    # tsdiffana
    screen_res['ts_res'] = time_slice_diffs(data,
                                            time_axis=time_axis,
                                            slice_axis=slice_axis)
    return screen_res


def write_screen_res(res, out_path, out_root,
                     out_img_ext='.nii',
                     pcnt_var_thresh=0.1):
    ''' Write results from ``screen`` to disk as images

    Parameters
    ----------
    res : dict
       output from ``screen`` function
    out_path : str
       directory to which to write output images
    out_root : str
       part of filename between image-specific prefix and image-specific
       extension to use for writing images
    out_img_ext : str, optional
       extension (identifying image type) to which to write volume
       images.  Default is '.nii'
    pcnt_var_thresh : float, optional
       threshold below which we do not plot percent variance explained
       by components; default is 0.1.  This removes the long tail from
       percent variance plots.

    Returns
    -------
    None
    '''
    import matplotlib.pyplot as plt
    # save volume images
    for key in ('mean', 'min', 'max', 'std', 'pca'):
        fname = pjoin(out_path, '%s_%s%s' % (key,
                                             out_root,
                                             out_img_ext))
        save_image(res[key], fname)
    # plot, save component time courses and some tsdiffana stuff
    pca_axis = res['pca_res']['axis']
    n_comp = res['pca_res']['basis_projections'].shape[pca_axis]
    vectors = res['pca_res']['basis_vectors']
    pcnt_var = res['pca_res']['pcnt_var']
    np.savez(pjoin(out_path, 'vectors_components_%s.npz' % out_root),
             basis_vectors=vectors,
             pcnt_var=pcnt_var,
             volume_means=res['ts_res']['volume_means'],
             slice_mean_diff2=res['ts_res']['slice_mean_diff2'],
            )
    plt.figure()
    for c in range(n_comp):
        plt.subplot(n_comp, 1, c+1)
        plt.plot(vectors[:,c])
        plt.axis('tight')
    plt.suptitle(out_root + ': PCA basis vectors')
    plt.savefig(pjoin(out_path, 'components_%s.png' % out_root))
    # plot percent variance
    plt.figure()
    plt.plot(pcnt_var[pcnt_var >= pcnt_var_thresh])
    plt.axis('tight')
    plt.suptitle(out_root + ': PCA percent variance')
    plt.savefig(pjoin(out_path, 'pcnt_var_%s.png' % out_root))    
    # plot tsdiffana
    plt.figure()
    axes = [plt.subplot(4, 1, i+1) for i in range(4)]
    plot_tsdiffs(res['ts_res'], axes)
    plt.suptitle(out_root + ': tsdiffana')
    plt.savefig(pjoin(out_path, 'tsdiff_%s.png' % out_root))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Motion correction / motion correction with slice timing

Routines implementing motion correction and motion correction combined with
slice-timing.

See:

Roche, Alexis (2011) A four-dimensional registration algorithm with application
to joint correction of motion and slice timing in fMRI. *Medical Imaging, IEEE
Transactions on*;  30:1546--1554
"""
from __future__ import absolute_import
from __future__ import print_function

import os
import warnings

import numpy as np

from ...externals.six import string_types

from nibabel.affines import apply_affine

from ...fixes.nibabel import io_orientation
from ...io.nibcompat import get_header
from ...core.image.image_spaces import (make_xyz_image,
                                        xyz_affine,
                                        as_xyz_image)
from ..slicetiming import timefuncs
from .affine import Rigid, Affine
from .optimizer import configure_optimizer, use_derivatives
from .type_check import (check_type, check_type_and_shape)
from ._registration import (_cspline_transform,
                            _cspline_sample3d,
                            _cspline_sample4d)


VERBOSE = os.environ.get('NIPY_DEBUG_PRINT', False)
INTERLEAVED = None
XTOL = 1e-5
FTOL = 1e-5
GTOL = 1e-5
STEPSIZE = 1e-6
SMALL = 1e-20
MAXITER = 64
MAXFUN = None


def interp_slice_times(Z, slice_times, tr):
    Z = np.asarray(Z)
    nslices = len(slice_times)
    aux = np.asarray(list(slice_times) + [slice_times[0] + tr])
    Zf = np.floor(Z).astype('int')
    w = Z - Zf
    Zal = Zf % nslices
    Za = Zal + w
    ret = (1 - w) * aux[Zal] + w * aux[Zal + 1]
    ret += (Z - Za)
    return ret


def scanner_coords(xyz, affine, from_world, to_world):
    Tv = np.dot(from_world, np.dot(affine, to_world))
    XYZ = apply_affine(Tv, xyz)
    return XYZ[:, 0], XYZ[:, 1], XYZ[:, 2]


def make_grid(dims, subsampling=(1, 1, 1), borders=(0, 0, 0)):
    slices = [slice(b, d - b, s)\
                  for d, s, b in zip(dims, subsampling, borders)]
    xyz = np.mgrid[slices]
    xyz = np.rollaxis(xyz, 0, 4)
    xyz = np.reshape(xyz, [np.prod(xyz.shape[0:-1]), 3])
    return xyz


def guess_slice_axis_and_direction(slice_info, affine):
    if slice_info is None:
        orient = io_orientation(affine)
        slice_axis = int(np.where(orient[:, 0] == 2)[0])
        slice_direction = int(orient[slice_axis, 1])
    else:
        slice_axis = int(slice_info[0])
        slice_direction = int(slice_info[1])
    return slice_axis, slice_direction

def tr_from_header(images):
    """ Return the TR from the header of an image or list of images.

    Parameters
    ----------
    images : image or list of images
      Single or multiple input 4d images representing one or
      several sessions.

    Returns
    -------
    float
      Repetition time, as specified in NIfTI header.

    Raises
    ------
    ValueError
      if the TR between the images is inconsistent.
    """
    if not isinstance(images, list):
        images = [images]
    images_tr = None
    for image in images:
        tr = get_header(image).get_zooms()[3]
        if images_tr is None:
            images_tr = tr
        if tr != images_tr:
            raise ValueError('TR inconsistent between images.')
    return images_tr

class Image4d(object):
    """
    Class to represent a sequence of 3d scans (possibly acquired on a
    slice-by-slice basis).

    Object remains empty until the data array is actually loaded in memory.

    Parameters
    ----------
      data : nd array or proxy (function that actually gets the array)
    """
    def __init__(self, data, affine, tr, slice_times, slice_info=None):
        """
        Configure fMRI acquisition time parameters.
        """
        self.affine = np.asarray(affine)
        self.tr = float(tr)

        # guess the slice axis and direction (z-axis)
        self.slice_axis, self.slice_direction =\
            guess_slice_axis_and_direction(slice_info, self.affine)

        # unformatted parameters
        self._slice_times = slice_times

        if isinstance(data, np.ndarray):
            self._data = data
            self._shape = data.shape
            self._get_data = None
            self._init_timing_parameters()
        else:
            self._data = None
            self._shape = None
            self._get_data = data

    def _load_data(self):
        self._data = self._get_data()
        self._shape = self._data.shape
        self._init_timing_parameters()

    def get_data(self):
        if self._data is None:
            self._load_data()
        return self._data

    def get_shape(self):
        if self._shape is None:
            self._load_data()
        return self._shape

    def _init_timing_parameters(self):
        # Number of slices
        nslices = self.get_shape()[self.slice_axis]
        self.nslices = nslices
        # Set slice times
        if isinstance(self._slice_times, (int, float)):
            # If a single value is provided, assume synchronous slices
            self.slice_times = np.zeros(nslices)
            self.slice_times.fill(self._slice_times)
        else:
            # Verify correctness of provided slice times
            if not len(self._slice_times) == nslices:
                raise ValueError(
                    "Incorrect slice times were provided. There are %d "
                    "slices in the volume, `slice_times` argument has length %d"
                    % (nslices, len(self._slice_times)))
            self.slice_times = np.asarray(self._slice_times)
        # Check that slice times are smaller than repetition time
        if np.max(self.slice_times) > self.tr:
            raise ValueError("slice times should be smaller than repetition time")

    def z_to_slice(self, z):
        """
        Account for the fact that slices may be stored in reverse
        order wrt the scanner coordinate system convention (slice 0 ==
        bottom of the head)
        """
        if self.slice_direction < 0:
            return self.nslices - 1 - z
        else:
            return z

    def scanner_time(self, zv, t):
        """
        tv = scanner_time(zv, t)
        zv, tv are grid coordinates; t is an actual time value.
        """
        corr = interp_slice_times(self.z_to_slice(zv),
                                  self.slice_times,
                                  self.tr)
        return (t - corr) / self.tr

    def free_data(self):
        if self._get_data is not None:
            self._data = None



class Realign4dAlgorithm(object):

    def __init__(self,
                 im4d,
                 affine_class=Rigid,
                 transforms=None,
                 time_interp=True,
                 subsampling=(1, 1, 1),
                 refscan=0,
                 borders=(1, 1, 1),
                 optimizer='ncg',
                 optimize_template=True,
                 xtol=XTOL,
                 ftol=FTOL,
                 gtol=GTOL,
                 stepsize=STEPSIZE,
                 maxiter=MAXITER,
                 maxfun=MAXFUN):

        # Check arguments
        check_type_and_shape(subsampling, int, 3)
        check_type(refscan, int, accept_none=True)
        check_type_and_shape(borders, int, 3)
        check_type(xtol, float)
        check_type(ftol, float)
        check_type(gtol, float)
        check_type(stepsize, float)
        check_type(maxiter, int)
        check_type(maxfun, int, accept_none=True)

        # Get dimensional parameters
        self.dims = im4d.get_shape()
        self.nscans = self.dims[3]
        # Reduce borders if spatial image dimension too small to avoid
        # getting an empty volume of interest
        borders = [min(b, d/2 - (not d%2)) for (b, d) in zip(borders, self.dims[0:3])]
        self.xyz = make_grid(self.dims[0:3], subsampling, borders)
        masksize = self.xyz.shape[0]
        self.data = np.zeros([masksize, self.nscans], dtype='double')

        # Initialize space/time transformation parameters
        self.affine = im4d.affine
        self.inv_affine = np.linalg.inv(self.affine)
        if transforms is None:
            self.transforms = [affine_class() for scan in range(self.nscans)]
        else:
            self.transforms = transforms

        # Compute the 4d cubic spline transform
        self.time_interp = time_interp
        if time_interp:
            self.timestamps = im4d.tr * np.arange(self.nscans)
            self.scanner_time = im4d.scanner_time
            self.cbspline = _cspline_transform(im4d.get_data())
        else:
            self.cbspline = np.zeros(self.dims, dtype='double')
            for t in range(self.dims[3]):
                self.cbspline[:, :, :, t] =\
                    _cspline_transform(im4d.get_data()[:, :, :, t])

        # The reference scan conventionally defines the head
        # coordinate system
        self.optimize_template = optimize_template
        if not optimize_template and refscan is None:
            self.refscan = 0
        else:
            self.refscan = refscan

        # Set the minimization method
        self.set_fmin(optimizer, stepsize,
                      xtol=xtol,
                      ftol=ftol,
                      gtol=gtol,
                      maxiter=maxiter,
                      maxfun=maxfun)

        # Auxiliary array for realignment estimation
        self._res = np.zeros(masksize, dtype='double')
        self._res0 = np.zeros(masksize, dtype='double')
        self._aux = np.zeros(masksize, dtype='double')
        self.A = np.zeros((masksize, self.transforms[0].param.size),
                          dtype='double')
        self._pc = None

    def resample(self, t):
        """
        Resample a particular time frame on the (sub-sampled) working
        grid.

        x,y,z,t are "head" grid coordinates
        X,Y,Z,T are "scanner" grid coordinates
        """
        X, Y, Z = scanner_coords(self.xyz, self.transforms[t].as_affine(),
                                 self.inv_affine, self.affine)
        if self.time_interp:
            T = self.scanner_time(Z, self.timestamps[t])
            _cspline_sample4d(self.data[:, t],
                              self.cbspline,
                              X, Y, Z, T,
                              mx='reflect',
                              my='reflect',
                              mz='reflect',
                              mt='reflect')
        else:
            _cspline_sample3d(self.data[:, t],
                              self.cbspline[:, :, :, t],
                              X, Y, Z,
                              mx='reflect',
                              my='reflect',
                              mz='reflect')

    def resample_full_data(self):
        if VERBOSE:
            print('Gridding...')
        xyz = make_grid(self.dims[0:3])
        res = np.zeros(self.dims)
        for t in range(self.nscans):
            if VERBOSE:
                print('Fully resampling scan %d/%d' % (t + 1, self.nscans))
            X, Y, Z = scanner_coords(xyz, self.transforms[t].as_affine(),
                                     self.inv_affine, self.affine)
            if self.time_interp:
                T = self.scanner_time(Z, self.timestamps[t])
                _cspline_sample4d(res[:, :, :, t],
                                  self.cbspline,
                                  X, Y, Z, T,
                                  mt='nearest')
            else:
                _cspline_sample3d(res[:, :, :, t],
                                  self.cbspline[:, :, :, t],
                                  X, Y, Z)
        return res

    def set_fmin(self, optimizer, stepsize, **kwargs):
        """
        Return the minimization function
        """
        self.stepsize = stepsize
        self.optimizer = optimizer
        self.optimizer_kwargs = kwargs
        self.optimizer_kwargs.setdefault('xtol', XTOL)
        self.optimizer_kwargs.setdefault('ftol', FTOL)
        self.optimizer_kwargs.setdefault('gtol', GTOL)
        self.optimizer_kwargs.setdefault('maxiter', MAXITER)
        self.optimizer_kwargs.setdefault('maxfun', MAXFUN)
        self.use_derivatives = use_derivatives(self.optimizer)

    def init_instant_motion(self, t):
        """
        Pre-compute and cache some constants (at fixed time) for
        repeated computations of the alignment energy.

        The idea is to decompose the average temporal variance via:

        V = (n-1)/n V* + (n-1)/n^2 (x-m*)^2

        with x the considered volume at time t, and m* the mean of all
        resampled volumes but x. Only the second term is variable when

        one volumes while the others are fixed. A similar
        decomposition is used for the global variance, so we end up
        with:

        V/V0 = [nV* + (x-m*)^2] / [nV0* + (x-m0*)^2]
        """
        fixed = list(range(self.nscans))
        fixed.remove(t)
        aux = self.data[:, fixed]
        if self.optimize_template:
            self.mu = np.mean(aux, 1)
        self.offset = self.nscans * np.mean((aux.T - self.mu) ** 2)
        self.mu0 = np.mean(aux)
        self.offset0 = self.nscans * np.mean((aux - self.mu0) ** 2)
        self._t = t
        self._pc = None

    def set_transform(self, t, pc):
        self.transforms[t].param = pc
        self.resample(t)

    def _init_energy(self, pc):
        if pc is self._pc:
            return
        self.set_transform(self._t, pc)
        self._pc = pc
        self._res[:] = self.data[:, self._t] - self.mu[:]
        self._V = np.maximum(self.offset + np.mean(self._res ** 2), SMALL)
        self._res0[:] = self.data[:, self._t] - self.mu0
        self._V0 = np.maximum(self.offset0 + np.mean(self._res0 ** 2), SMALL)

        if self.use_derivatives:
            # linearize the data wrt the transform parameters
            # use the auxiliary array to save the current resampled data
            self._aux[:] = self.data[:, self._t]
            basis = np.eye(6)
            for j in range(pc.size):
                self.set_transform(self._t, pc + self.stepsize * basis[j])
                self.A[:, j] = (self.data[:, self._t] - self._aux)\
                    / self.stepsize
            self.transforms[self._t].param = pc
            self.data[:, self._t] = self._aux[:]
            # pre-compute gradient and hessian of numerator and
            # denominator
            c = 2 / float(self.data.shape[0])
            self._dV = c * np.dot(self.A.T, self._res)
            self._dV0 = c * np.dot(self.A.T, self._res0)
            self._H = c * np.dot(self.A.T, self.A)

    def _energy(self):
        """
        The alignment energy is defined as the log-ratio between the
        average temporal variance in the sequence and the global
        spatio-temporal variance.
        """
        return np.log(self._V / self._V0)

    def _energy_gradient(self):
        return self._dV / self._V - self._dV0 / self._V0

    def _energy_hessian(self):
        return (1 / self._V - 1 / self._V0) * self._H\
            - np.dot(self._dV, self._dV.T) / np.maximum(self._V ** 2, SMALL)\
            + np.dot(self._dV0, self._dV0.T) / np.maximum(self._V0 ** 2, SMALL)

    def estimate_instant_motion(self, t):
        """
        Estimate motion parameters at a particular time.
        """
        if VERBOSE:
            print('Estimating motion at time frame %d/%d...'
                  % (t + 1, self.nscans))

        def f(pc):
            self._init_energy(pc)
            return self._energy()

        def fprime(pc):
            self._init_energy(pc)
            return self._energy_gradient()

        def fhess(pc):
            self._init_energy(pc)
            return self._energy_hessian()

        self.init_instant_motion(t)
        fmin, args, kwargs =\
            configure_optimizer(self.optimizer,
                                fprime=fprime,
                                fhess=fhess,
                                **self.optimizer_kwargs)

        # With scipy >= 0.9, some scipy minimization functions like
        # fmin_bfgs may crash due to the subroutine
        # `scalar_search_armijo` returning None as a stepsize when
        # unhappy about the objective function. This seems to have the
        # potential to occur in groupwise registration when using
        # strong image subsampling, i.e. at the coarser levels of the
        # multiscale pyramid. To avoid crashes, we insert a try/catch
        # instruction.
        try:
            pc = fmin(f, self.transforms[t].param, disp=VERBOSE,
                      *args, **kwargs)
            self.set_transform(t, pc)
        except:
            warnings.warn('Minimization failed')

    def estimate_motion(self):
        """
        Optimize motion parameters for the whole sequence. All the
        time frames are initially resampled according to the current
        space/time transformation, the parameters of which are further
        optimized sequentially.
        """
        for t in range(self.nscans):
            if VERBOSE:
                print('Resampling scan %d/%d' % (t + 1, self.nscans))
            self.resample(t)

        # Set the template as the reference scan (will be overwritten
        # if template is to be optimized)
        if not hasattr(self, 'template'):
            self.mu = self.data[:, self.refscan].copy()
        for t in range(self.nscans):
            self.estimate_instant_motion(t)
            if VERBOSE:
                print(self.transforms[t])

    def align_to_refscan(self):
        """
        The `motion_estimate` method aligns scans with an online
        template so that spatial transforms map some average head
        space to the scanner space. To conventionally redefine the
        head space as being aligned with some reference scan, we need
        to right compose each head_average-to-scanner transform with
        the refscan's 'to head_average' transform.
        """
        if self.refscan is None:
            return
        Tref_inv = self.transforms[self.refscan].inv()
        for t in range(self.nscans):
            self.transforms[t] = (self.transforms[t]).compose(Tref_inv)


def resample4d(im4d, transforms, time_interp=True):
    """
    Resample a 4D image according to the specified sequence of spatial
    transforms, using either 4D interpolation if `time_interp` is True
    and 3D interpolation otherwise.
    """
    r = Realign4dAlgorithm(im4d, transforms=transforms,
                           time_interp=time_interp)
    res = r.resample_full_data()
    im4d.free_data()
    return res


def adjust_subsampling(speedup, dims):
    dims = np.array(dims)
    aux = np.maximum(speedup * dims / np.prod(dims) ** (1 / 3.), [1, 1, 1])
    return aux.astype('int')


def single_run_realign4d(im4d,
                         affine_class=Rigid,
                         time_interp=True,
                         loops=5,
                         speedup=5,
                         refscan=0,
                         borders=(1, 1, 1),
                         optimizer='ncg',
                         xtol=XTOL,
                         ftol=FTOL,
                         gtol=GTOL,
                         stepsize=STEPSIZE,
                         maxiter=MAXITER,
                         maxfun=MAXFUN):
    """
    Realign a single run in space and time.

    Parameters
    ----------
    im4d : Image4d instance

    speedup : int or sequence
      If a sequence, implement a multi-scale realignment
    """
    if not type(loops) in (list, tuple, np.array):
        loops = [loops]
    repeats = len(loops)

    def format_arg(x):
        if not type(x) in (list, tuple, np.array):
            x = [x for i in range(repeats)]
        else:
            if not len(x) == repeats:
                raise ValueError('inconsistent length in arguments')
        return x

    speedup = format_arg(speedup)
    optimizer = format_arg(optimizer)
    xtol = format_arg(xtol)
    ftol = format_arg(ftol)
    gtol = format_arg(gtol)
    stepsize = format_arg(stepsize)
    maxiter = format_arg(maxiter)
    maxfun = format_arg(maxfun)

    transforms = None
    opt_params = zip(loops, speedup, optimizer,
                     xtol, ftol, gtol,
                     stepsize, maxiter, maxfun)

    for loops_, speedup_, optimizer_, xtol_, ftol_, gtol_,\
            stepsize_, maxiter_, maxfun_ in opt_params:
        subsampling = adjust_subsampling(speedup_, im4d.get_shape()[0:3])

        r = Realign4dAlgorithm(im4d,
                               transforms=transforms,
                               affine_class=affine_class,
                               time_interp=time_interp,
                               subsampling=subsampling,
                               refscan=refscan,
                               borders=borders,
                               optimizer=optimizer_,
                               xtol=xtol_,
                               ftol=ftol_,
                               gtol=gtol_,
                               stepsize=stepsize_,
                               maxiter=maxiter_,
                               maxfun=maxfun_)

        for loop in range(loops_):
            r.estimate_motion()

        r.align_to_refscan()
        transforms = r.transforms
        im4d.free_data()

    return transforms


def realign4d(runs,
              affine_class=Rigid,
              time_interp=True,
              align_runs=True,
              loops=5,
              between_loops=5,
              speedup=5,
              refscan=0,
              borders=(1, 1, 1),
              optimizer='ncg',
              xtol=XTOL,
              ftol=FTOL,
              gtol=GTOL,
              stepsize=STEPSIZE,
              maxiter=MAXITER,
              maxfun=MAXFUN):
    """
    Parameters
    ----------
    runs : list of Image4d objects

    Returns
    -------
    transforms : list
                 nested list of rigid transformations


    transforms map an 'ideal' 4d grid (conventionally aligned with the
    first scan of the first run) to the 'acquisition' 4d grid for each
    run
    """

    # Single-session case
    if not type(runs) in (list, tuple, np.array):
        runs = [runs]
    nruns = len(runs)
    if nruns == 1:
        align_runs = False

    # Correct motion and slice timing in each sequence separately
    transforms = [single_run_realign4d(run,
                                       affine_class=affine_class,
                                       time_interp=time_interp,
                                       loops=loops,
                                       speedup=speedup,
                                       refscan=refscan,
                                       borders=borders,
                                       optimizer=optimizer,
                                       xtol=xtol,
                                       ftol=ftol,
                                       gtol=gtol,
                                       stepsize=stepsize,
                                       maxiter=maxiter,
                                       maxfun=maxfun) for run in runs]

    if not align_runs:
        return transforms, transforms, None

    # Correct between-session motion using the mean image of each
    # corrected run, and creating a fake time series with no temporal
    # smoothness. If the runs have different affines, a correction is
    # applied to the transforms associated with each run (except for
    # the first run) so that all images included in the fake series
    # have the same affine, namely that of the first run.
    is_same_affine = lambda a1, a2: np.max(np.abs(a1 - a2)) < 1e-5
    mean_img_shape = list(runs[0].get_shape()[0:3]) + [nruns]
    mean_img_data = np.zeros(mean_img_shape)
    for i in range(nruns):
        if is_same_affine(runs[0].affine, runs[i].affine):
            transforms_i = transforms[i]
        else:
            runs[i].affine = runs[0].affine
            aff_corr = Affine(np.dot(runs[0].affine,
                                     np.linalg.inv(runs[i].affine)))
            transforms_i = [aff_corr.compose(Affine(t.as_affine()))\
                                for t in transforms[i]]
        corr_run = resample4d(runs[i], transforms=transforms_i,
                              time_interp=time_interp)
        mean_img_data[..., i] = corr_run.mean(3)
    del corr_run

    mean_img = Image4d(mean_img_data, affine=runs[0].affine,
                       tr=1.0, slice_times=0)
    transfo_mean = single_run_realign4d(mean_img,
                                        affine_class=affine_class,
                                        time_interp=False,
                                        loops=between_loops,
                                        speedup=speedup,
                                        borders=borders,
                                        optimizer=optimizer,
                                        xtol=xtol,
                                        ftol=ftol,
                                        gtol=gtol,
                                        stepsize=stepsize,
                                        maxiter=maxiter,
                                        maxfun=maxfun)

    # Compose transformations for each run
    ctransforms = [None for i in range(nruns)]
    for i in range(nruns):
        ctransforms[i] = [t.compose(transfo_mean[i]) for t in transforms[i]]
    return ctransforms, transforms, transfo_mean


class Realign4d(object):

    def __init__(self, images, tr, slice_times=None, slice_info=None,
                 affine_class=Rigid):
        """
        Spatiotemporal realignment class for series of 3D images.

        The algorithm performs simultaneous motion and slice timing
        correction for fMRI series or other data where slices are not
        acquired simultaneously.

        Parameters
        ----------
        images : image or list of images
          Single or multiple input 4d images representing one or
          several sessions.

        tr : float
          Inter-scan repetition time, i.e. the time elapsed between
          two consecutive scans. The unit in which `tr` is given is
          arbitrary although it needs to be consistent with the
          `slice_times` argument.

        slice_times : None or array-like
          If None, slices are assumed to be acquired simultaneously
          hence no slice timing correction is performed. If
          array-like, then the slice acquisition times.

        slice_info : None or tuple, optional
          None, or a tuple with slice axis as the first element and
          direction as the second, for instance (2, 1).  If None, then
          guess the slice axis, and direction, as the closest to the z
          axis, as estimated from the affine.
        """
        self._init(images, tr, slice_times, slice_info, affine_class)

    def _init(self, images, tr, slice_times, slice_info, affine_class):
        """
        Generic initialization method.
        """
        if slice_times is None:
            tr = 1.0
            slice_times = 0.0
            time_interp = False
        else:
            time_interp = True
        if not isinstance(images, (list, tuple, np.ndarray)):
            images = [images]
        if tr is None:
            raise ValueError('Repetition time cannot be None.')
        if tr == 0:
            raise ValueError('Repetition time cannot be zero.')
        self.affine_class = affine_class
        self.slice_times = slice_times
        self.tr = tr
        self._runs = []
        # Note that, the affine of each run may be different. This is
        # the case, for instance, if the subject exits the scanner
        # inbetween sessions.
        for im in images:
            xyz_img = as_xyz_image(im)
            self._runs.append(Image4d(xyz_img.get_data,
                                      xyz_affine(xyz_img),
                                      tr,
                                      slice_times=slice_times,
                                      slice_info=slice_info))
        self._transforms = [None for run in self._runs]
        self._within_run_transforms = [None for run in self._runs]
        self._mean_transforms = [None for run in self._runs]
        self._time_interp = time_interp

    def estimate(self,
                 loops=5,
                 between_loops=None,
                 align_runs=True,
                 speedup=5,
                 refscan=0,
                 borders=(1, 1, 1),
                 optimizer='ncg',
                 xtol=XTOL,
                 ftol=FTOL,
                 gtol=GTOL,
                 stepsize=STEPSIZE,
                 maxiter=MAXITER,
                 maxfun=MAXFUN):
        """Estimate motion parameters.

        Parameters
        ----------
        loops : int or sequence of ints
            Determines the number of iterations performed to realign
            scans within each run for each pass defined by the
            ``speedup`` argument. For instance, setting ``speedup`` ==
            (5,2) and ``loops`` == (5,1) means that 5 iterations are
            performed in a first pass where scans are subsampled by an
            isotropic factor 5, followed by one iteration where scans
            are subsampled by a factor 2.
        between_loops : None, int or sequence of ints
            Similar to ``loops`` for between-run motion
            estimation. Determines the number of iterations used to
            realign scans across runs, a procedure similar to
            within-run realignment that uses the mean images from each
            run. If None, assumed to be the same as ``loops``.
            The setting used in the experiments described in Roche,
            IEEE TMI 2011, was: ``speedup`` = (5, 2), ``loops`` = (5,
            1) and ``between_loops`` = (5, 1).
        align_runs : bool
            Determines whether between-run motion is estimated or
            not. If False, the ``between_loops`` argument is ignored.
        speedup: int or sequence of ints
            Determines an isotropic sub-sampling factor, or a sequence
            of such factors, applied to the scans to perform motion
            estimation. If a sequence, several estimation passes are
            applied.
        refscan : None or int
            Defines the number of the scan used as the reference
            coordinate system for each run. If None, a reference
            coordinate system is defined internally that does not
            correspond to any particular scan. Note that the
            coordinate system associated with the first run is always
        borders : sequence of ints
            Should be of length 3. Determines the field of view for
            motion estimation in terms of the number of slices at each
            extremity of the reference grid that are ignored for
            motion parameter estimation. For instance,
            ``borders``==(1,1,1) means that the realignment cost
            function will not take into account voxels located in the
            first and last axial/sagittal/coronal slices in the
            reference grid. Please note that this choice only affects
            parameter estimation but does not affect image resampling
            in any way, see ``resample`` method.
        optimizer : str
            Defines the optimization method. One of 'simplex',
            'powell', 'cg', 'ncg', 'bfgs' and 'steepest'.
        xtol : float
            Tolerance on variations of transformation parameters to
            test numerical convergence.
        ftol : float
            Tolerance on variations of the intensity comparison metric
            to test numerical convergence.
        gtol : float
            Tolerance on the gradient of the intensity comparison
            metric to test numerical convergence. Applicable to
            optimizers 'cg', 'ncg', 'bfgs' and 'steepest'.
        stepsize : float
            Step size to approximate the gradient and Hessian of the
            intensity comparison metric w.r.t. transformation
            parameters. Applicable to optimizers 'cg', 'ncg', 'bfgs'
            and 'steepest'.
        maxiter : int
            Maximum number of iterations in optimization.
        maxfun : int
            Maximum number of function evaluations in maxfun.
        """
        if between_loops is None:
            between_loops = loops
        t = realign4d(self._runs,
                      affine_class=self.affine_class,
                      time_interp=self._time_interp,
                      align_runs=align_runs,
                      loops=loops,
                      between_loops=between_loops,
                      speedup=speedup,
                      refscan=refscan,
                      borders=borders,
                      optimizer=optimizer,
                      xtol=xtol,
                      ftol=ftol,
                      gtol=gtol,
                      stepsize=stepsize,
                      maxiter=maxiter,
                      maxfun=maxfun)
        self._transforms, self._within_run_transforms,\
            self._mean_transforms = t

    def resample(self, r=None, align_runs=True):
        """
        Return the resampled run number r as a 4d nipy-like
        image. Returns all runs as a list of images if r is None.
        """
        if align_runs:
            transforms = self._transforms
        else:
            transforms = self._within_run_transforms
        runs = range(len(self._runs))
        if r is None:
            data = [resample4d(self._runs[r], transforms=transforms[r],
                               time_interp=self._time_interp) for r in runs]
            return [make_xyz_image(data[r], self._runs[r].affine, 'scanner')
                    for r in runs]
        else:
            data = resample4d(self._runs[r], transforms=transforms[r],
                              time_interp=self._time_interp)
            return make_xyz_image(data, self._runs[r].affine, 'scanner')


class SpaceTimeRealign(Realign4d):

    def __init__(self, images, tr, slice_times, slice_info,
                 affine_class=Rigid):
        """ Spatiotemporal realignment class for fMRI series.

        This class gives a high-level interface to :class:`Realign4d`

        Parameters
        ----------
        images : image or list of images
            Single or multiple input 4d images representing one or several fMRI
            runs.
        tr : None or float or "header-allow-1.0"
            Inter-scan repetition time in seconds, i.e. the time elapsed between
            two consecutive scans. If None, an attempt is made to read the TR
            from the header, but an exception is thrown for values 0 or 1. A
            value of "header-allow-1.0" will signal to accept a header TR of 1.
        slice_times : str or callable or array-like
            If str, one of the function names in ``SLICETIME_FUNCTIONS``
            dictionary from :mod:`nipy.algorithms.slicetiming.timefuncs`.  If
            callable, a function taking two parameters: ``n_slices`` and ``tr``
            (number of slices in the images, inter-scan repetition time in
            seconds). This function returns a vector of times of slice
            acquisition $t_i$ for each slice $i$ in the volumes.  See
            :mod:`nipy.algorithms.slicetiming.timefuncs` for a collection of
            functions for common slice acquisition schemes. If array-like, then
            should be a slice time vector as above.
        slice_info : int or length 2 sequence
            If int, the axis in `images` that is the slice axis.  In a 4D image,
            this will often be axis = 2.  If a 2 sequence, then elements are
            ``(slice_axis, slice_direction)``, where ``slice_axis`` is the slice
            axis in the image as above, and ``slice_direction`` is 1 if the
            slices were acquired slice 0 first, slice -1 last, or -1 if acquired
            slice -1 first, slice 0 last.  If `slice_info` is an int, assume
            ``slice_direction`` == 1.
        affine_class : ``Affine`` class, optional
            transformation class to use to calculate transformations between
            the volumes. Default is :class:``Rigid``
        """
        if tr is None:
            tr = tr_from_header(images)
            if tr == 1:
                raise ValueError('A TR of 1 was found in the header. '
                    'This value often stands in for an unknown TR. '
                    'Please specify TR explicitly. Alternatively '
                    'consider setting TR to "header-allow-1.0".')
        elif tr == "header-allow-1.0":
            tr = tr_from_header(images)
        if tr == 0:
            raise ValueError('Repetition time cannot be zero.')
        if slice_times is None:
            raise ValueError("slice_times must be set for space/time "
                             "registration; use SpaceRealign for space-only "
                             "registration")
        if slice_info is None:
            raise ValueError("slice_info cannot be None")
        try:
            len(slice_info)
        except TypeError:
            # Presumably an int
            slice_axis = slice_info
            slice_info = (slice_axis, 1)
        else: # sequence
            slice_axis, slice_direction = slice_info
        if type(images) in (list, tuple):
            n_slices = images[0].shape[slice_axis]
        else:
            n_slices = images.shape[slice_axis]
        if isinstance(slice_times, string_types):
            slice_times = timefuncs.SLICETIME_FUNCTIONS[slice_times]
        if hasattr(slice_times, '__call__'):
            slice_times = slice_times(n_slices, tr)
        self._init(images, tr, slice_times, slice_info, affine_class)


class SpaceRealign(Realign4d):

    def __init__(self, images, affine_class=Rigid):
        """ Spatial registration of time series with no time interpolation

        Parameters
        ----------
        images : image or list of images
            Single or multiple input 4d images representing one or several fMRI
            runs.
        affine_class : ``Affine`` class, optional
            transformation class to use to calculate transformations between
            the volumes. Default is :class:``Rigid``
        """
        self._init(images, 1., None, None, affine_class)


class FmriRealign4d(Realign4d):

    def __init__(self, images, slice_order=None,
                 tr=None, tr_slices=None, start=0.0,
                 interleaved=None, time_interp=None,
                 slice_times=None,
                 affine_class=Rigid, slice_info=None):
        """
        Spatiotemporal realignment class for fMRI series. This class
        is similar to `Realign4d` but provides a more flexible API for
        initialization in order to make it easier to declare slice
        acquisition times for standard sequences.

        Warning: this class is deprecated; please use :class:`SpaceTimeRealign`
        instead.

        Parameters
        ----------
        images : image or list of images
          Single or multiple input 4d images representing one or
          several fMRI runs.

        slice_order : str or array-like
          If str, one of {'ascending', 'descending'}. If array-like,
          then the order in which the slices were collected in
          time. For instance, the following represents an ascending
          contiguous sequence:

          slice_order = [0, 1, 2, ...]

          Note that `slice_order` differs from the argument used
          e.g. in the SPM slice timing routine in that it maps spatial
          slice positions to slice times. It is a mapping from space
          to time, while SPM conventionally uses the reverse mapping
          from time to space. For example, for an interleaved sequence
          with 10 slices, where we acquired slice 0 (in space) first,
          then slice 2 (in space) etc, `slice_order` would be [0, 5,
          1, 6, 2, 7, 3, 8, 4, 9]

          Using `slice_order` assumes that the inter-slice acquisition
          time is constant throughout acquisition. If this is not the
          case, use the `slice_times` argument instead and leave
          `slice_order` to None.

        tr : float
          Inter-scan repetition time, i.e. the time elapsed between
          two consecutive scans. The unit in which `tr` is given is
          arbitrary although it needs to be consistent with the
          `tr_slices` and `start` arguments if provided. If None, `tr`
          is computed internally assuming a regular slice acquisition
          scheme.

        tr_slices : float
          Inter-slice repetition time, same as `tr` for slices. If
          None, acquisition is assumed regular and `tr_slices` is set
          to `tr` divided by the number of slices.

        start : float
          Starting acquisition time (time of the first acquired slice)
          respective to the time origin for resampling. `start` is
          assumed to be given in the same unit as `tr`. Setting
          `start=0` means that the resampled data will be synchronous
          with the first acquired slice. Setting `start=-tr/2` means
          that the resampled data will be synchronous with the slice
          acquired at half repetition time.

        time_interp: bool
          Tells whether time interpolation is used or not within the
          realignment algorithm. If False, slices are considered to be
          acquired all at the same time, thus no slice timing
          correction will be performed.

        interleaved : bool
          Deprecated argument.

          Tells whether slice acquisition order is interleaved in a
          certain sense. Setting `interleaved` to True or False will
          trigger an error unless `slice_order` is 'ascending' or
          'descending' and `slice_times` is None.

          If slice_order=='ascending' and interleaved==True, the
          assumed slice order is (assuming 10 slices):

          [0, 5, 1, 6, 2, 7, 3, 8, 4, 9]

          If slice_order=='descending' and interleaved==True, the
          assumed slice order is:

          [9, 4, 8, 3, 7, 2, 6, 1, 5, 0]

          WARNING: given that there exist other types of interleaved
          acquisitions depending on scanner settings and
          manufacturers, you should refrain from using the
          `interleaved` keyword argument unless you are sure what you
          are doing. It is generally safer to explicitly input
          `slice_order` or `slice_times`.

        slice_times : None, str or array-like

          This argument can be used instead of `slice_order`,
          `tr_slices`, `start` and `time_interp` altogether.

          If None, slices are assumed to be acquired simultaneously
          hence no slice timing correction is performed. If
          array-like, then `slice_times` gives the slice acquisition
          times along the slice axis in units that are consistent with
          the provided `tr`.

          Generally speaking, the following holds for sequences with
          constant inter-slice repetition time `tr_slices`:

          `slice_times` = `start` + `tr_slices` * `slice_order`

          For other sequences such as, e.g., sequences with
          simultaneously acquired slices, it is necessary to input
          `slice_times` explicitly along with `tr`.

        slice_info : None or tuple, optional
          None, or a tuple with slice axis as the first element and
          direction as the second, for instance (2, 1). If None, then
          the slice axis and direction are guessed from the first
          run's affine assuming that slices are collected along the
          closest axis to the z-axis. This means that we assume by
          default an axial acquisition with slice axis pointing from
          bottom to top of the head.
        """
        warnings.warn('Please use SpaceTimeRealign instead of this class; '
                      'We will soon remove this class',
                      FutureWarning,
                      stacklevel=2)
        # if slice_times not None, make sure that parameters redundant
        # with slice times all have their default value
        if slice_times is not None:
            if slice_order is not None \
                    or tr_slices is not None\
                    or start != 0.0 \
                    or time_interp is not None\
                    or interleaved is not None:
                raise ValueError('Attempting to set both `slice_times` '
                                 'and other arguments redundant with it')
            if tr is None:
                if len(slice_times) > 1:
                    tr = slice_times[-1] + slice_times[1] - 2 * slice_times[0]
                else:
                    tr = 2 * slice_times[0]
                warnings.warn('No `tr` entered. Assuming regular acquisition'
                              ' with tr=%f' % tr)
        # case where slice_time is None
        else:
            # assume regular slice acquisition, therefore tr is
            # arbitrary
            if tr is None:
                tr = 1.0
            # if no slice order provided, assume synchronous slices
            if slice_order is None:
                if not time_interp == False:
                    raise ValueError('Slice order is requested '
                                     'with time interpolation switched on')
                slice_times = 0.0
            else:
                # if slice_order is a key word, replace it with the
                # appropriate array of slice indices
                if slice_order in ('ascending', 'descending'):
                    if isinstance(images, (list, tuple, np.array)):
                        xyz_img = as_xyz_image(images[0])
                    else:
                        xyz_img = as_xyz_image(images)

                    slice_axis, _ = guess_slice_axis_and_direction(
                        slice_info, xyz_affine(xyz_img))
                    nslices = xyz_img.shape[slice_axis]
                    if interleaved:
                        warnings.warn('`interleaved` keyword argument is '
                                      'deprecated',
                                      FutureWarning,
                                      stacklevel=2)
                        aux = np.argsort(list(range(0, nslices, 2)) +
                                         list(range(1, nslices, 2)))
                    else:
                        aux = np.arange(nslices)
                    if slice_order == 'descending':
                        aux = aux[::-1]
                    slice_order = aux
                # if slice_order is provided explicitly, issue a
                # warning and make sure interleaved is set to None
                else:
                    warnings.warn('Please make sure you are NOT using '
                                  'SPM-style slice order declaration')
                    if interleaved is not None:
                        raise ValueError('`interleaved` should be None when '
                                         'providing explicit slice order')
                    slice_order = np.asarray(slice_order)
                if tr_slices is None:
                    tr_slices = float(tr) / float(len(slice_order))
                if start is None:
                    start = 0.0
                slice_times = start + tr_slices * slice_order

        self._init(images, tr, slice_times, slice_info, affine_class)



""" Helper functions for constructing design regressors
"""
from __future__ import division

import numpy as np


def dct_ii_basis(volume_times, order=None, normcols=False):
    """ DCT II basis up to order `order`

    See: https://en.wikipedia.org/wiki/Discrete_cosine_transform#DCT-II

    By default, basis not normalized to length 1, and therefore, basis is not
    orthogonal.  Normalize basis with `normcols` keyword argument.

    Parameters
    ----------
    volume_times : array-like
        Times of acquisition of each volume.  Must be regular and continuous
        otherwise we raise an error.
    order : None or int, optional
        Order of DCT-II basis.  If None, return full basis set.
    normcols : bool, optional
        If True, normalize columns to length 1, so return orthogonal
        `dct_basis`.

    Returns
    -------
    dct_basis : array
        Shape ``(len(volume_times), order)`` array with DCT-II basis up to
        order `order`.

    Raises
    ------
    ValueError
        If difference between successive `volume_times` values is not constant
        over the 1D array.
    """
    N = len(volume_times)
    if order is None:
        order = N
    if not np.allclose(np.diff(np.diff(volume_times)), 0):
        raise ValueError("DCT basis assumes continuous regular sampling")
    n = np.arange(N)
    cycle = np.pi * (n + 0.5) / N
    dct_basis = np.zeros((N, order))
    for k in range(0, order):
        dct_basis[:, k] = np.cos(cycle * k)
    if normcols:  # Set column lengths to 1
        lengths = np.ones(order) * np.sqrt(N / 2.)
        lengths[0:1] = np.sqrt(N)  # Allow order=0
        dct_basis /= lengths
    return dct_basis


def dct_ii_cut_basis(volume_times, cut_period):
    """DCT-II regressors with periods >= `cut_period`

    See: http://en.wikipedia.org/wiki/Discrete_cosine_transform#DCT-II

    Parameters
    ----------
    volume_times : array-like
        Times of acquisition of each volume.  Must be regular and continuous
        otherwise we raise an error.
    cut_period: float
        Cut period (wavelength) of the low-pass filter (in time units).

    Returns
    -------
    cdrift: array shape (n_scans, n_drifts)
        DCT-II drifts plus a constant regressor in the final column.  Constant
        regressor always present, regardless of `cut_period`.
    """
    N = len(volume_times)
    hfcut = 1./ cut_period
    dt = volume_times[1] - volume_times[0]
    # Such that hfcut = 1/(2*dt) yields N
    order = int(np.floor(2 * N * hfcut * dt))
    # Always return constant column
    if order == 0:
        return np.ones((N, 1))
    basis = np.ones((N, order))
    basis[:, :-1] = dct_ii_basis(volume_times, order, normcols=True)[:, 1:]
    return basis



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The main routine of this package that aims at performing the
extraction of ROIs from multisubject dataset using the localization
and activation strength of extracted regions.

This has been published in:
- Thirion et al. High level group analysis of FMRI data based on
Dirichlet process mixture models, IPMI 2007
- Thirion et al.
Accurate Definition of Brain Regions Position Through the
Functional Landmark Approach, MICCAI 2010

Author : Bertrand Thirion, 2006-2013
"""
from __future__ import absolute_import

import numpy as np
import scipy.stats as st

from .structural_bfls import build_landmarks
from nipy.algorithms.graph import wgraph_from_coo_matrix
from ...algorithms.statistics.empirical_pvalue import \
    NormalEmpiricalNull, three_classes_GMM_fit, gamma_gaussian_fit
from .hroi import HROI_as_discrete_domain_blobs

####################################################################
# Ancillary functions
####################################################################


def _stat_to_proba(test, learn=None, method='prior', alpha=0.01, verbose=0):
    """Convert a set of statistics to posterior probabilities of  being
    generated under H0

    Parameters
    ----------
    test: array of shape(n_samples),
           data that is assessed
    learn: array of shape(n_samples), optional,
           data to learn a mixture model
           defaults to learn
    method: {'gauss_mixture', 'emp_null', 'gam_gauss', 'prior'}, optional,
           'gauss_mixture' A Gaussian Mixture Model is used
           'emp_null' a null mode is fitted to test
           'gam_gauss' a Gamma-Gaussian mixture is used
           'prior' a hard-coded function is used
    alpha: float in the [0,1] range, optional,
           parameter that yields the prior probability that a region is active
           should be chosen close to 0
    verbose: int, optional,
             verbosity mode

    Returns
    -------
    posterior_null: array of shape(n_samples)
                    an estimation of the probability that the observation
                    is generated under the null
    """
    if method == 'gauss_mixture':
        prior_strength = 100
        fixed_scale = True
        posterior_probas = three_classes_GMM_fit(
            learn, test, alpha, prior_strength, verbose, fixed_scale)
        posterior_null = posterior_probas[:, 1]
    elif method == 'emp_null':
        enn = NormalEmpiricalNull(learn)
        enn.learn()
        posterior_null = np.reshape(enn.fdr(test), np.size(test))
    elif method == 'gam_gauss':
        posterior_probas = gamma_gaussian_fit(learn, test, verbose)
        posterior_null = posterior_probas[:, 1]
    elif method == 'prior':
        y0 = st.norm.pdf(test)
        shape_, scale_ = 3., 2.
        y1 = st.gamma.pdf(test, shape_, scale=scale_)
        posterior_null = np.ravel(
            (1 - alpha) * y0 / (alpha * y1 + (1 - alpha) * y0))
    else:
        raise ValueError('Unknown method')
    return posterior_null


def _compute_individual_regions(domain, stats, threshold=3.0, smin=5,
                               method='gauss_mixture'):
    """ Compute the individual regions that are real activation candidates

    Parameters
    ----------
    domain : StructuredDomain instance,
          generic descriptor of the space domain
    stats: an array of shape (n_voxels, n_subjects)
           the multi-subject statistical maps
    threshold: float, optional
           first level threshold
    smin: int, optional
          minimal size of the regions to validate them
    method: {'gauss_mixture', 'emp_null', 'gam_gauss', 'prior'}, optional,
           'gauss_mixture' A Gaussian Mixture Model is used
           'emp_null' a null mode is fitted to test
           'gam_gauss' a Gamma-Gaussian mixture is used
           'prior' a hard-coded function is used

    Returns
    -------
    hrois: list of nipy.labs.spatial_models.hroi.HierrachicalROI instances
           that represent individual ROIs
           let nr be the number of terminal regions across subjects
    prior_h0: array of shape (nr),
              the mixture-based prior probability
              that the terminal regions are false positives
    subjects: array of shape (nr),
              the subject index associated with the terminal regions
    coords: array of shape (nr, coord.shape[1]),
            the coordinates of the of the terminal regions

    Fixme
    -----
    Should allow for subject specific domains
    """
    hrois = []
    coords = []
    prior_h0 = []
    subjects = []
    n_subjects = stats.shape[1]
    nvox = stats.shape[0]

    for subject in range(n_subjects):
        # description in terms of blobs
        stats_ = np.reshape(stats[:, subject], (nvox, 1))
        hroi = HROI_as_discrete_domain_blobs(
            domain, stats_, threshold=threshold, smin=smin)

        if hroi is not None and hroi.k > 0:
            # get the leave regions (i.e. the local maxima)
            leaves = [hroi.select_id(id) for id in hroi.get_leaves_id()]

            # get the region mean statistical value
            mean_val = hroi.representative_feature('signal', 'weighted mean')
            mean_val = mean_val[leaves]

            # get the regions position
            mean_pos = np.asarray(
                [np.mean(coord, 0) for coord in hroi.get_coord()])
            hroi.set_roi_feature('position', mean_pos)
            coords.append(mean_pos[leaves])

            # compute the prior proba of being null
            learning_set = np.squeeze(stats_[stats_ != 0])
            prior_h0.append(_stat_to_proba(mean_val, learning_set, method))
            subjects.append(subject * np.ones(mean_val.size).astype(np.int))
        else:
            subjects.append([])
            prior_h0.append([])
            coords.append(np.empty((0, domain.dim)))
        hrois.append(hroi)

    prior_h0 = np.concatenate(prior_h0)
    subjects = np.concatenate(subjects)
    coords = np.concatenate(coords)
    return hrois, prior_h0, subjects, coords


def _dpmm(coords, alpha, null_density, dof, prior_precision, prior_h0,
          subjects, sampling_coords=None, n_iter=1000, burnin=100,
          co_clust=False):
    """Apply the dpmm analysis to compute clusters from regions coordinates
    """
    from nipy.algorithms.clustering.imm import MixedIMM

    dim = coords.shape[1]
    migmm = MixedIMM(alpha, dim)
    migmm.set_priors(coords)
    migmm.set_constant_densities(
        null_dens=null_density, prior_dens=null_density)
    migmm._prior_dof = dof
    migmm._prior_scale = np.diag(prior_precision[0] / dof)
    migmm._inv_prior_scale_ = [np.diag(dof * 1. / (prior_precision[0]))]
    migmm.sample(coords, null_class_proba=prior_h0, niter=burnin, init=False,
                 kfold=subjects)

    # sampling
    like, pproba, co_clustering = migmm.sample(
        coords, null_class_proba=prior_h0, niter=n_iter, kfold=subjects,
        sampling_points=sampling_coords, co_clustering=True)

    if co_clust:
        return like, 1 - pproba, co_clustering
    else:
        return like, 1 - pproba


def _update_hroi_labels(hrois, new_labels):
    """Update the labels of the hroisusing new_labels"""
    for subject in range(len(hrois)):
        if hrois[subject].k > 0:
            us = hrois[subject].get_roi_feature('label')
            us[us > - 1] = new_labels[us[us > - 1]]
            hrois[subject].set_roi_feature('label', us)


def _bsa_dpmm(hrois, prior_h0, subjects, coords, sigma, prevalence_pval,
             prevalence_threshold, dof=10, alpha=.5, n_iter=1000, burnin=100,
             algorithm='density'):
    """ Estimation of the population level model of activation density using
    dpmm and inference

    Parameters
    ----------
    hrois: list of nipy.labs.spatial_models.hroi.HierarchicalROI instances
           representing individual ROIs
           Let nr be the number of terminal regions across subjects
    prior_h0: array of shape (nr)
              mixture-based prior probability
              that the terminal regions are true positives
    subjects: array of shape (nr)
              subject index associated with the terminal regions
    coords: array of shape (nr, coord.shape[1])
            coordinates of the of the terminal regions
    sigma: float > 0,
           expected cluster scatter in the common space in units of coord
    prevalence_pval: float in the [0,1] interval, optional
                     p-value of the prevalence test
    prevalence_threshold: float in the rannge [0,nsubj]
                         null hypothesis on region prevalence
    dof: float > 0, optional,
         degrees of freedom of the prior
    alpha: float > 0, optional,
           creation parameter of the DPMM
    niter: int, optional,
           number of iterations of the DPMM
    burnin: int, optional,
            number of iterations of the DPMM
    algorithm: {'density', 'co_occurrence'}, optional,
               algorithm used in the DPMM inference

    Returns
    -------
    landmarks: instance of sbf.LandmarkRegions
               that describes the ROIs found in inter-subject inference
               If no such thing can be defined landmarks is set to None
    hrois: List of nipy.labs.spatial_models.hroi.HierarchicalROI instances
           representing individual ROIs
    """
    from nipy.algorithms.graph.field import field_from_coo_matrix_and_data
    domain = hrois[0].domain
    n_subjects = len(hrois)

    landmarks = None
    density = np.zeros(domain.size)
    if len(subjects) < 1:
        return landmarks, hrois

    null_density = 1. / domain.local_volume.sum()

    # prepare the DPMM
    dim = domain.em_dim
    prior_precision = 1. / (sigma ** 2) * np.ones((1, dim))

    # n_iter = number of iterations to estimate density
    if algorithm == 'density':
        density, post_proba = _dpmm(
            coords, alpha, null_density, dof, prior_precision, prior_h0,
            subjects, domain.coord, n_iter=n_iter, burnin=burnin)
            # associate labels with coords
        Fbeta = field_from_coo_matrix_and_data(domain.topology, density)
        _, label = Fbeta.custom_watershed(0, null_density)
        midx = np.array([np.argmin(np.sum((domain.coord - coord_) ** 2, 1))
                         for coord_ in coords])
        components = label[midx]
    elif algorithm == 'co-occurrence':
        post_proba, density, co_clustering = _dpmm(
            coords, alpha, null_density, dof, prior_precision, prior_h0,
            subjects,  n_iter=n_iter, burnin=burnin, co_clust=True)
        contingency_graph = wgraph_from_coo_matrix(co_clustering)
        if contingency_graph.E > 0:
            contingency_graph.remove_edges(contingency_graph.weights > .5)

        components = contingency_graph.cc()
        components[density < null_density] = components.max() + 1 +\
            np.arange(np.sum(density < null_density))
    else:
        raise ValueError('Unknown algorithm')

    # append some information to the hroi in each subject
    for subject in range(n_subjects):
        bfs = hrois[subject]
        if bfs is None:
            continue
        if bfs.k == 0:
            bfs.set_roi_feature('label', np.array([]))
            continue

        leaves_pos = [bfs.select_id(k) for k in bfs.get_leaves_id()]
        # save posterior proba
        post_proba_ = np.zeros(bfs.k)
        post_proba_[leaves_pos] = post_proba[subjects == subject]
        bfs.set_roi_feature('posterior_proba', post_proba_)

        # save prior proba
        prior_proba = np.zeros(bfs.k)
        prior_proba[leaves_pos] = 1 - prior_h0[subjects == subject]
        bfs.set_roi_feature('prior_proba', prior_proba)

        # assign labels to ROIs
        roi_label = - np.ones(bfs.k).astype(np.int)
        roi_label[leaves_pos] = components[subjects == subject]
        # when parent regions has similarly labelled children,
        # include it also
        roi_label = bfs.make_forest().propagate_upward(roi_label)
        bfs.set_roi_feature('label', roi_label)

    # derive the group-level landmarks
    # with a threshold on the number of subjects
    # that are represented in each one
    landmarks, new_labels = build_landmarks(
        domain, coords, subjects, np.array(components), 1 - prior_h0,
        prevalence_pval, prevalence_threshold, sigma)

    # relabel the regions
    _update_hroi_labels(hrois, new_labels)

    return landmarks, hrois


###########################################################################
# Main function
###########################################################################


def compute_landmarks(
    domain, stats, sigma, prevalence_pval=0.5, prevalence_threshold=0,
    threshold=3.0, smin=5, method='prior', algorithm='density', n_iter=1000,
    burnin=100):
    """ Compute the  Bayesian Structural Activation patterns

    Parameters
    ----------
    domain: StructuredDomain instance,
            Description of the spatial context of the data
    stats: array of shape (nbnodes, subjects):
           the multi-subject statistical maps
    sigma: float > 0:
           expected cluster std in the common space in units of coord
    prevalence_pval: float in the [0,1] interval, optional
                     posterior significance threshold
    prevalence_threshold: float, optional,
                          reference threshold for the prevalence value
    threshold: float, optional,
               first level threshold
    smin: int, optional,
          minimal size of the regions to validate them
    method: {'gauss_mixture', 'emp_null', 'gam_gauss', 'prior'}, optional,
           'gauss_mixture' A Gaussian Mixture Model is used
           'emp_null' a null mode is fitted to test
           'gam_gauss' a Gamma-Gaussian mixture is used
           'prior' a hard-coded function is used
    algorithm: string, one of ['density', 'co-occurrence'], optional
               method used to compute the landmarks
    niter: int, optional,
           number of iterations of the DPMM
    burnin: int, optional,
            number of iterations of the DPMM

    Returns
    -------
    landmarks: Instance of sbf.LandmarkRegions or None,
               Describes the ROIs found in inter-subject inference
               None if nothing can be defined
    hrois: list of  nipy.labs.spatial_models.hroi.Nroi instances
           representing individual ROIs
    """
    hrois, prior_h0, subjects, coords = _compute_individual_regions(
        domain, stats, threshold, smin, method)

    landmarks, hrois = _bsa_dpmm(
        hrois, prior_h0, subjects, coords, sigma, prevalence_pval,
        prevalence_threshold, algorithm=algorithm, n_iter=n_iter,
        burnin=burnin)

    return landmarks, hrois
